# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import numpy as np
from chainladder.utils.sparse import sp
from sklearn.base import BaseEstimator


class WeightedRegression(BaseEstimator):
    """ Helper class that fits a system of regression equations
        as a closed-form solution.  This greatly speeds up
        the implementation of the Mack stochastic properties.
    """

    def __init__(self, axis=None, thru_orig=False, xp=None):
        self.axis = axis
        self.thru_orig = thru_orig
        self.xp = xp

    def infer_x_w(self):
        xp = self.xp
        if self.w is None:
            self.w = xp.ones(self.y.shape)
        if self.x is None:
            self.x = xp.cumsum(xp.ones(self.y.shape), self.axis)
        return self

    def fit(self, X, y=None, sample_weight=None):
        self.x = X
        self.y = y
        self.w = sample_weight
        if self.x is None:
            self.infer_x_w()
        if self.thru_orig:
            self._fit_OLS_thru_orig()
        else:
            self._fit_OLS()
        return self

    def _fit_OLS(self):
        """ Given a set of w, x, y, and an axis, this Function
            returns OLS slope and intercept.
            TODO:
                Make this work with n_periods = 1 without numpy warning.
        """
        from chainladder.utils.utility_functions import num_to_nan

        w, x, y, axis = self.w.copy(), self.x.copy(), self.y.copy(), self.axis
        xp = self.xp
        if xp != sp:
            x[w == 0] = xp.nan
            y[w == 0] = xp.nan
        else:
            w2 = w.copy()
            w2 = sp(data=w2.data, coords=w2.coords, fill_value=sp.nan, shape=w2.shape)
            x, y = x * w2, y * w2
        slope = num_to_nan(
            xp.nansum(w * x * y, axis) - xp.nansum(x * w, axis) * xp.nanmean(y, axis)
        ) / num_to_nan(
            xp.nansum(w * x * x, axis) - xp.nanmean(x, axis) * xp.nansum(w * x, axis)
        )
        intercept = xp.nanmean(y, axis) - slope * xp.nanmean(x, axis)
        self.slope_ = slope[..., None]
        self.intercept_ = intercept[..., None]
        return self

    def _fit_OLS_thru_orig(self):
        from chainladder.utils.utility_functions import num_to_nan

        w, x, y, axis = self.w, self.x, self.y, self.axis
        xp = self.xp
        d = num_to_nan(xp.nansum((y * 0 + 1) * w * x * x, axis))
        coef = num_to_nan(xp.nansum(w * x * y, axis)) / d
        fitted_value = xp.repeat(xp.expand_dims(coef, axis), x.shape[axis], axis)
        fitted_value = fitted_value * x * (y * 0 + 1)
        residual = (y - fitted_value) * xp.sqrt(w)
        wss_residual = xp.nansum(residual ** 2, axis)
        mse_denom = xp.nansum((y * 0 + 1) * (w != 0), axis) - 1
        mse_denom = num_to_nan(mse_denom)
        mse = wss_residual / mse_denom
        std_err = xp.sqrt(mse / d)
        std_err = std_err[..., None]
        coef = coef[..., None]
        sigma = xp.sqrt(mse)[..., None]
        self.slope_ = coef
        self.sigma_ = sigma
        self.std_err_ = std_err
        return self

    def sigma_fill(self, interpolation):
        """ This Function is designed to take an array of sigmas and does log-
            linear extrapolation where n_obs=1 and sigma cannot be calculated.
        """
        if interpolation == "log-linear":
            self.sigma_ = self.loglinear_interpolation(self.sigma_)
        if interpolation == "mack":
            self.sigma_ = self.mack_interpolation(self.sigma_)
        return self

    def std_err_fill(self):
        """currently handled in development.py which doesn't feel right"""
        return self

    def loglinear_interpolation(self, y):
        """ Use Cases: generally for filling in last element of sigma_
        """
        from chainladder.utils.utility_functions import num_to_nan

        xp = self.xp
        ly = y.copy()
        ly = xp.log(xp.where(ly==0, 1e-320, ly))
        w = xp.nan_to_num(ly * 0 + 1)
        reg = WeightedRegression(self.axis, False, xp=xp).fit(None, ly, w)
        slope, intercept = reg.slope_, reg.intercept_
        fill_ = xp.exp(reg.x * slope + intercept) * (1 - w)
        out = xp.nan_to_num(y) + xp.nan_to_num(fill_)
        return num_to_nan(out)

    def mack_interpolation(self, y):
        """ Use Mack's approximation to fill last element of sigma_ which is the
            same as loglinear extrapolation using the preceding two element to
            the missing value. This function needs a recursive definition...
        """
        from chainladder.utils.utility_functions import num_to_nan

        xp = self.xp
        ly = y.copy()
        ly = xp.where(ly == 0, 1e-320, ly)
        w = xp.nan_to_num(ly * 0 + 1)
        slicer_n, slicer_d, slicer_a = (
            ([slice(None)] * 4),
            ([slice(None)] * 4),
            ([slice(None)] * 4),
        )
        slicer_n[self.axis], slicer_d[self.axis], slicer_a[self.axis] = (
            slice(1, -1, 1),
            slice(0, -2, 1),
            slice(0, 2, 1),
        )
        slicer_n, slicer_d, slicer_a = (
            tuple(slicer_n),
            tuple(slicer_d),
            tuple(slicer_a),
        )
        fill_ = xp.sqrt(
            abs(
                xp.minimum(
                    (ly[slicer_n] ** 4 / ly[slicer_d] ** 2),
                    xp.minimum(ly[slicer_d] ** 2, ly[slicer_n] ** 2),
                )
            )
        )
        fill_ = xp.concatenate((w[slicer_a], xp.nan_to_num(fill_)), axis=self.axis) * (
            1 - w
        )
        out = xp.nan_to_num(ly) + fill_
        return num_to_nan(out)



from chainladder.development.base import DevelopmentBase  # noqa (API import)
from chainladder.development.development import Development  # noqa (API import)
from chainladder.development.munich import MunichAdjustment  # noqa (API import)
from chainladder.development.incremental import IncrementalAdditive  # noqa (API import)
from chainladder.development.constant import DevelopmentConstant  # noqa (API import)
from chainladder.development.clark import ClarkLDF  # noqa (API import)
from chainladder.development.outstanding import CaseOutstanding  # noqa (API import)
from chainladder.development.learning import DevelopmentML  # noqa (API import)
from chainladder.development.glm import TweedieGLM  # noqa (API import)
from chainladder.development.barnzehn import BarnettZehnwirth  # noqa (API import)



import pytest
import numpy as np
import chainladder as cl


@pytest.fixture
def atol():
    return 1e-5


data = ["RAA", "ABC", "GenIns", "MW2008", "MW2014"]


@pytest.mark.parametrize("data", data)
def test_benktander_to_chainladder(data, atol):
    tri = cl.load_sample(data)
    a = cl.Chainladder().fit(tri).ibnr_
    b = cl.Benktander(apriori=0.8, n_iters=255).fit(tri, sample_weight=a).ibnr_
    xp = tri.get_array_module()
    assert xp.allclose(xp.nan_to_num(a.values), xp.nan_to_num(b.values), atol=atol)


def test_bf_eq_cl_when_using_cl_apriori(qtr):
    cl_ult = cl.Chainladder().fit(qtr).ultimate_
    bf_ult = cl.BornhuetterFerguson().fit(qtr, sample_weight=cl_ult).ultimate_
    xp = cl_ult.get_array_module()
    assert xp.allclose(cl_ult.values, bf_ult.values, atol=1e-5)


def test_different_backends(clrd):
    clrd = clrd[["CumPaidLoss", "EarnedPremDIR"]]
    clrd = clrd[clrd["LOB"] == "wkcomp"]
    model = cl.BornhuetterFerguson().fit(
        clrd["CumPaidLoss"].sum().set_backend("numpy"),
        sample_weight=clrd["EarnedPremDIR"].sum().latest_diagonal.set_backend("numpy"),
    )
    assert (
        abs(
            (
                model.predict(
                    clrd["CumPaidLoss"].set_backend("sparse"),
                    sample_weight=clrd["EarnedPremDIR"].latest_diagonal.set_backend(
                        "sparse"
                    ),
                ).ibnr_.sum()
                - model.ibnr_
            ).sum()
        )
        < 1
    )



import chainladder as cl
import numpy as np


def test_bs_sample(raa):
    tri = raa
    a = (
        cl.Development()
        .fit(cl.BootstrapODPSample(n_sims=40000).fit_transform(tri).mean())
        .ldf_
    )
    b = cl.Development().fit_transform(tri).ldf_
    assert tri.get_array_module().all(abs(((a - b) / b).values) < 0.005)

def test_bs_multiple_cols():
    assert cl.BootstrapODPSample().fit_transform(
        cl.load_sample('berqsherm').iloc[0]).shape == (1000, 4, 8, 8)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import numpy as np
import pandas as pd
from chainladder.methods import Chainladder


class MackChainladder(Chainladder):
    """ Basic stochastic chainladder method popularized by Thomas Mack

    Parameters
    ----------
    None

    Attributes
    ----------
    X_:
        returns **X**
    ultimate_:
        The ultimate losses per the method
    ibnr_:
        The IBNR per the method
    full_expectation_:
        The ultimates back-filled to each development period in **X** replacing
        the known data
    full_triangle_:
        The ultimates back-filled to each development period in **X** retaining
        the known data
    summary_:
        summary of the model
    full_std_err_:
        The full standard error
    total_process_risk_:
        The total process error
    total_parameter_risk_:
        The total parameter error
    mack_std_err_:
        The total prediction error by origin period
    total_mack_std_err_:
        The total prediction error across all origin periods
    """

    def fit(self, X, y=None, sample_weight=None):
        """Fit the model with X.

        Parameters
        ----------
        X: Triangle-like
            Data to which the model will be applied.
        y: Ignored
        sample_weight: Ignored

        Returns
        -------
        self: object
            Returns the instance itself.
        """
        super().fit(X, y, sample_weight)
        if "sigma_" not in self.X_:
            raise ValueError("Triangle not compatible with MackChainladder")
        # Caching full_triangle_ for fit as it is called a lot
        self.X_._full_triangle_ = self.full_triangle_
        self.parameter_risk_ = self._mack_recursion("parameter_risk_", self.X_)
        self.process_risk_ = self._mack_recursion("process_risk_", self.X_)
        self.total_parameter_risk_ = self._mack_recursion(
            "total_parameter_risk_", self.X_
        )
        del self.X_._full_triangle_
        self.process_variance_ = self._include_process_variance()
        return self

    def predict(self, X, sample_weight=None):
        X_new = super().predict(X, sample_weight)
        X_new.sigma_ = getattr(X_new, "sigma_", self.X_.sigma_)
        X_new.std_err_ = getattr(X_new, "std_err_", self.X_.std_err_)
        X_new.average_ = getattr(X_new, "average_", self.X_.average_)
        X_new.w_ = getattr(X_new, "w_", self.X_.w_)
        X_new._full_triangle_ = X_new.full_triangle_
        X_new.parameter_risk_ = self._mack_recursion("parameter_risk_", X_new)
        X_new.process_risk_ = self._mack_recursion("process_risk_", X_new)
        X_new.total_process_risk_ = (X_new.process_risk_ ** 2).sum(axis="origin").sqrt()
        X_new.total_parameter_risk_ = self._mack_recursion(
            "total_parameter_risk_", X_new
        )
        X_new.full_std_err_ = self._get_full_std_err_(X_new)
        X_new.total_mack_std_err_ = self._get_total_mack_std_err_(X_new)
        X_new.mack_std_err_ = (
            X_new.parameter_risk_ ** 2 + X_new.process_risk_ ** 2
        ).sqrt()
        del X_new._full_triangle_
        return X_new

    @property
    def full_std_err_(self):
        return self._get_full_std_err_(self.X_)

    def _get_full_std_err_(self, X=None):
        from chainladder.utils.utility_functions import num_to_nan

        obj = X.copy()
        xp = obj.get_array_module()
        lxp = X.ldf_.get_array_module()
        full = getattr(X, "_full_triangle_", self.full_triangle_)
        avg = {"regression": 0, "volume": 1, "simple": 2}
        avg = [avg.get(item, item) for item in X.average_]
        val = xp.broadcast_to(xp.array(avg + [avg[-1]]), X.shape)
        weight = xp.sqrt(full.values[..., : len(X.ddims)] ** (2 - val))
        obj.values = X.sigma_.values / num_to_nan(weight)
        w = lxp.concatenate((X.w_, lxp.ones((1, 1, val.shape[2], 1))), 3)
        w[xp.isnan(w)] = 1
        obj.values = xp.nan_to_num(obj.values) * xp.array(w)
        obj.valuation_date = full.valuation_date
        obj._set_slicers()
        return obj

    @property
    def total_process_risk_(self):
        return (self.process_risk_ ** 2).sum(axis="origin").sqrt()

    def _mack_recursion(self, est, X=None):
        obj = X.copy()
        xp = obj.get_array_module()
        risk_arr = xp.zeros((*X.shape[:3], 1))
        if est == "total_parameter_risk_":
            nans = None
            risk_arr = risk_arr[..., 0:1, :]
            future_std_err = (
                X._full_triangle_ - X[X.valuation < X.valuation_date]
            ).iloc[:, :, :, : X.shape[3]] * X.std_err_.values
            t1_t = xp.nan_to_num(future_std_err.sum("origin").values)
            obj.odims = obj.odims[0:1]
        else:
            nans = xp.nan_to_num(X.nan_triangle[None, None])
            nans = 1 - xp.concatenate((nans, xp.zeros((1, 1, X.shape[2], 1))), 3)
            full_tri = X._full_triangle_.values[..., : len(X.ddims)]
            if est == "parameter_risk_":
                t1_t = xp.nan_to_num(full_tri) * obj.std_err_.values
            else:
                t1_t = xp.nan_to_num(full_tri) * self._get_full_std_err_(X).values
        extend = X.ldf_.shape[-1] - X.shape[-1] + 1
        ldf = X.ldf_.values[..., : len(X.ddims) - 1]
        tail = X.cdf_.values[..., -extend : -extend + 1]
        ldf = xp.array(X.ldf_.get_array_module().concatenate((ldf, tail), -1))
        # Recursive Mack Formula
        for i in range(t1_t.shape[-1]):
            t1 = t1_t[..., i : i + 1] ** 2
            t2 = (ldf[..., i : i + 1] * risk_arr[..., i : i + 1]) ** 2
            t_tot = xp.sqrt(t1 + t2)
            if nans is not None:
                t_tot = t_tot * nans[..., i + 1 : i + 2]
            risk_arr = xp.concatenate((risk_arr, xp.nan_to_num(t_tot)), 3)
        obj.values = risk_arr
        obj.ddims = X._full_triangle_.ddims[list(range(X.shape[-1])) + [-1]]
        obj.valuation_date = X._full_triangle_.valuation_date
        obj._set_slicers()
        return obj

    @property
    def mack_std_err_(self):
        return (self.parameter_risk_ ** 2 + self.process_risk_ ** 2).sqrt()

    @property
    def total_mack_std_err_(self):
        return self._get_total_mack_std_err_(self)

    def _get_total_mack_std_err_(self, obj):
        obj = obj.total_process_risk_ ** 2 + obj.total_parameter_risk_ ** 2
        if obj.array_backend == "sparse":
            out = obj.set_backend("numpy").sqrt().values[..., 0, -1]
        else:
            out = obj.sqrt().values[..., 0, -1]
        return pd.DataFrame(out, index=obj.index, columns=obj.columns)

    @property
    def summary_(self):
        # This might be better as a dataframe
        obj = self.ultimate_.copy()
        cols = (
            self.X_.latest_diagonal.values,
            self.ibnr_.values,
            self.ultimate_.values,
            self.mack_std_err_.values[..., -1:],
        )
        obj.values = obj.get_array_module().concatenate(cols, 3)
        obj.ddims = ["Latest", "IBNR", "Ultimate", "Mack Std Err"]
        obj._set_slicers()
        return obj



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import numpy as np

from chainladder import options
from chainladder.utils.sparse import sp

try:
    import cupy as cp

    cp.array([1])
    module = "cupy"
except:
    if options.ARRAY_BACKEND == "cupy":
        import warnings

        warnings.warn("Unable to load CuPY.  Using numpy instead.")
    import numpy as cp

    module = "numpy"


def nansum(a, *args, **kwargs):
    """ For cupy v0.6.0 compatibility """
    return cp.sum(cp.nan_to_num(a), *args, **kwargs)


def nanmean(a, *args, **kwargs):
    """ For cupy v0.6.0 compatibility """
    return cp.sum(cp.nan_to_num(a), *args, **kwargs) / cp.sum(
        ~cp.isnan(a), *args, **kwargs
    )


def nanmedian(a, *args, **kwargs):
    """ For cupy v0.6.0 compatibility """
    return cp.array(np.nanmedian(cp.asnumpy(a), *args, **kwargs))


def nanquantile(a, *args, **kwargs):
    """ For cupy v0.6.0 compatibility """
    return cp.array(np.nanquantile(cp.asnumpy(a), *args, **kwargs))


def unique(ar, axis=None, *args, **kwargs):
    """ For cupy v0.6.0 compatibility """
    return cp.array(np.unique(cp.asnumpy(ar), axis=axis, *args, **kwargs))


if module == "cupy":
    if int(cp.__version__.split(".")[0]) < 7:
        cp.nansum = nansum
        cp.nanmean = nanmean
        cp.nanmedian = nanmedian
        cp.unique = unique
    if int(cp.__version__.split(".")[0]) >= 7:
        cp.nanquantile = nanquantile



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from chainladder.methods import MethodBase
import numpy as np

class Benktander(MethodBase):
    """ The Benktander (or iterated Bornhuetter-Ferguson) IBNR model

    Parameters
    ----------
    apriori: float, optional (default=1.0)
        Multiplier for the sample_weight used in the Benktander method
        method. If sample_weight is already an apriori measure of ultimate,
        then use 1.0
    n_iters: int, optional (default=1)
        Number of iterations to use in the Benktander model.
    apriori_sigma: float, optional (default=0.0)
        Standard deviation of the apriori.  When used in conjunction with the
        bootstrap model, the model samples aprioris from a lognormal distribution
        using this argument as a standard deviation.
    random_state: int, RandomState instance or None, optional (default=None)
        Seed for sampling from the apriori distribution.  This is ignored when
        using as a deterministic method.
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by np.random.

    Attributes
    ----------
    ultimate_: Triangle
        The ultimate losses per the method
    ibnr_: Triangle
        The IBNR per the method
    """

    def __init__(self, apriori=1.0, n_iters=1, apriori_sigma=0, random_state=None):
        self.apriori = apriori
        self.n_iters = n_iters
        self.apriori_sigma = apriori_sigma
        self.random_state = random_state

    def fit(self, X, y=None, sample_weight=None):
        """Applies the Benktander technique to triangle **X**

        Parameters
        ----------
        X: Triangle
            Loss data to which the model will be applied.
        y: None
            Ignored
        sample_weight: Triangle
            Required exposure to be used in the calculation.

        Returns
        -------
        self: object
            Returns the instance itself.
        """
        if sample_weight is None:
            raise ValueError("sample_weight is required.")
        super().fit(X, y, sample_weight)
        self.expectation_ = self._get_benktander_aprioris(X, sample_weight)
        self.ultimate_ = self._get_ultimate(self.X_, self.expectation_)
        self.process_variance_ = self._include_process_variance()
        return self

    def predict(self, X, sample_weight=None):
        """Predicts the Benktander ultimate on a new triangle **X**

        Parameters
        ----------
        X: Triangle
            Loss data to which the model will be applied.
        sample_weight: Triangle
            Required exposure to be used in the calculation.

        Returns
        -------
        X_new: Triangle
            Loss data with Benktander ultimate applied
        """
        X_new = super().predict(X, sample_weight)
        X_new.expectation_ = self._get_benktander_aprioris(X, sample_weight)
        X_new.ultimate_ = self._get_ultimate(X_new, X_new.expectation_)
        X_new.n_iters = self.n_iters
        X_new.apriori = self.apriori
        return X_new

    def _get_benktander_aprioris(self, X, sample_weight):
        """ Private method to establish Benktander Apriori """
        xp = X.get_array_module()
        if self.apriori_sigma != 0:
            random_state = xp.random.RandomState(self.random_state)
            apriori = random_state.normal(self.apriori, self.apriori_sigma, X.shape[0])
            apriori = apriori.reshape(X.shape[0], -1)[..., None, None]
            apriori = sample_weight * apriori
            apriori.kdims = X.kdims
            apriori.key_labels = X.key_labels
        else:
            apriori = sample_weight * self.apriori
        apriori.columns = sample_weight.columns
        return apriori  

    def _get_ultimate(self, X, expectation):
        from chainladder.utils.utility_functions import num_to_nan
        if X.is_cumulative == False:
            ld = X.sum('development')
            ultimate = ld.val_to_dev()
        else:
            ld = X.latest_diagonal
            ultimate = X.copy()
        cdf = self._align_cdf(ultimate.val_to_dev(), expectation)
        backend = cdf.array_backend
        xp = cdf.get_array_module()
        cdf = cdf.sort_index()
        ld = ld.sort_index()
        expectation = expectation.sort_index()
        ultimate = ultimate.sort_index()
        cdf = (1 - 1 / num_to_nan(cdf.values))[None]
        exponents = xp.arange(self.n_iters + 1)
        exponents = xp.reshape(exponents, tuple([len(exponents)] + [1] * 4))
        cdf = cdf ** (((cdf + 1e-16) / (cdf + 1e-16) * exponents))
        cdf = xp.nan_to_num(cdf)
        a = xp.sum(cdf[:-1, ...], 0) * xp.nan_to_num(ld.set_backend(backend).values)
        b = cdf[-1, ...] * xp.nan_to_num(expectation.set_backend(backend).values)
        ultimate.values = num_to_nan(a + b)
        ultimate.array_backend = backend
        ultimate.ddims = self.cdf_.ddims[:ultimate.shape[-1]]
        return self._set_ult_attr(ultimate)



""" tails should store all tail methodologies
"""
from chainladder.tails.base import TailBase  # noqa (API import)
from chainladder.tails.constant import TailConstant  # noqa (API import)
from chainladder.tails.curve import TailCurve  # noqa (API import)
from chainladder.tails.bondy import TailBondy  # noqa (API import)
from chainladder.tails.clark import TailClark  # noqa (API import)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import numpy as np
import pandas as pd
import warnings
from chainladder.utils import WeightedRegression
from chainladder.development.base import DevelopmentBase


class Development(DevelopmentBase):
    """A Transformer that allows for basic loss development pattern selection.

    Parameters
    ----------
    n_periods: integer, optional (default = -1)
        number of origin periods to be used in the ldf average calculation. For
        all origin periods, set n_periods = -1
    average: string or float, optional (default = 'volume')
        type of averaging to use for ldf average calculation.  Options include
        'volume', 'simple', and 'regression'. If numeric values are supplied,
        then (2-average) in the style of Zehnwirth & Barnett is used
        for the exponent of the regression weights.
    sigma_interpolation: string optional (default = 'log-linear')
        Options include 'log-linear' and 'mack'
    drop: tuple or list of tuples
        Drops specific origin/development combination(s)
    drop_high: bool, int, list of bools, or list of ints (default = None)
        Drops highest (by rank) link ratio(s) from LDF calculation
        If a boolean variable is passed, drop_high is set to 1, dropping only the
        highest value
        Note that drop_high is performed after consideration of n_periods (if used)
    drop_low: bool, int, list of bools, or list of ints (default = None)
        Drops lowest (by rank) link ratio(s) from LDF calculation
        If a boolean variable is passed, drop_low is set to 1, dropping only the
        lowest value
        Note that drop_low is performed after consideration of n_periods (if used)
    drop_above: float or list of floats (default = numpy.inf)
        Drops all link ratio(s) above the given parameter from the LDF calculation
    drop_below: float or list of floats (default = 0.00)
        Drops all link ratio(s) below the given parameter from the LDF calculation
    preserve: int (default = 1)
        The minimum number of link ratio(s) required for LDF calculation
    drop_valuation: str or list of str (default = None)
        Drops specific valuation periods. str must be date convertible.
    fillna: float, (default = None)
        Used to fill in zero or nan values of an triangle with some non-zero
        amount.  When an link-ratio has zero as its denominator, it is automatically
        excluded from the ``ldf_`` calculation.  For the specific case of 'volume'
        averaging in a deterministic method, this may be reasonable.  For all other
        averages and stochastic methods, this assumption should be avoided.
    groupby:
        An option to group levels of the triangle index together for the purposes
        of estimating patterns.  If omitted, each level of the triangle
        index will receive its own patterns.


    Attributes
    ----------
    ldf_: Triangle
        The estimated loss development patterns
    cdf_: Triangle
        The estimated cumulative development patterns
    sigma_: Triangle
        Sigma of the ldf regression
    std_err_: Triangle
        Std_err of the ldf regression
    weight_: pandas.DataFrame
        The weight used in the ldf regression
    std_residuals_: Triangle
        A Triangle representing the weighted standardized residuals of the
        estimator as described in Barnett and Zehnwirth.

    """

    def __init__(
        self,
        n_periods=-1,
        average="volume",
        sigma_interpolation="log-linear",
        drop=None,
        drop_high=None,
        drop_low=None,
        preserve=1,
        drop_valuation=None,
        drop_above=np.inf,
        drop_below=0.00,
        fillna=None,
        groupby=None,
    ):
        self.n_periods = n_periods
        self.average = average
        self.sigma_interpolation = sigma_interpolation
        self.drop_high = drop_high
        self.drop_low = drop_low
        self.preserve = preserve
        self.drop_valuation = drop_valuation
        self.drop_above = drop_above
        self.drop_below = drop_below
        self.drop = drop
        self.fillna = fillna
        self.groupby = groupby

    def _validate_axis_assumption(self, parameter, axis):
        if callable(parameter):
            return axis.map(parameter).to_list()
        if type(parameter) in [int, str, float]:
            return [parameter] * len(axis)
        return parameter

    def fit(self, X, y=None, sample_weight=None):
        """Fit the model with X.

        Parameters
        ----------
        X : Triangle-like
            Set of LDFs to which the munich adjustment will be applied.
        y : None
            Ignored
        sample_weight : None
            Ignored

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        from chainladder.utils.utility_functions import num_to_nan

        # Validate inputs
        if X.is_cumulative is None:
            warnings.warn(
                "The is_cumulative property of your triangle is not set. This may result in undesirable behavior."
            )
        if X.is_cumulative == False:
            obj = self._set_fit_groups(X).incr_to_cum().val_to_dev().copy()
        else:
            obj = self._set_fit_groups(X).val_to_dev().copy()

        xp = obj.get_array_module()

        # Make sure it is a dev tri
        if type(obj.ddims) != np.ndarray:
            raise ValueError("Triangle must be expressed with development lags")

        # validate hyperparameters
        if self.fillna:
            tri_array = num_to_nan((obj + self.fillna).values)
        else:
            tri_array = num_to_nan(obj.values.copy())

        self.average_ = np.array(
            self._validate_axis_assumption(self.average, obj.development[:-1])
        )

        n_periods_ = self._validate_axis_assumption(
            self.n_periods, obj.development[:-1]
        )

        weight_dict = {"regression": 0, "volume": 1, "simple": 2}
        x, y = tri_array[..., :-1], tri_array[..., 1:]
        exponent = xp.array([weight_dict.get(item, item) for item in self.average_])
        exponent = xp.nan_to_num(exponent[None, None, None] * (y * 0 + 1))
        link_ratio = y / x

        self.w_ = self._assign_n_periods_weight(
            obj, n_periods_
        ) * self._drop_adjustment(obj, link_ratio)
        w = self.w_ / (x ** (exponent))
        params = WeightedRegression(axis=2, thru_orig=True, xp=xp).fit(x, y, w)
        if self.n_periods != 1:
            params = params.sigma_fill(self.sigma_interpolation)
        else:
            warnings.warn(
                "Setting n_periods=1 does not allow enough degrees "
                "of freedom to support calculation of all regression"
                " statistics.  Only LDFs have been calculated."
            )
        params.std_err_ = xp.nan_to_num(params.std_err_) + xp.nan_to_num(
            (1 - xp.nan_to_num(params.std_err_ * 0 + 1))
            * params.sigma_
            / xp.swapaxes(xp.sqrt(x ** (2 - exponent))[..., 0:1, :], -1, -2)
        )
        params = xp.concatenate((params.slope_, params.sigma_, params.std_err_), 3)
        params = xp.swapaxes(params, 2, 3)
        self.ldf_ = self._param_property(obj, params, 0)
        self.sigma_ = self._param_property(obj, params, 1)
        self.std_err_ = self._param_property(obj, params, 2)

        resid = -obj.iloc[..., :-1] * self.ldf_.values + obj.iloc[..., 1:].values

        std = xp.sqrt((1 / num_to_nan(w)) * (self.sigma_ ** 2).values)
        resid = resid / std
        self.std_residuals_ = resid[resid.valuation < obj.valuation_date]
        return self

    def transform(self, X):
        """If X and self are of different shapes, align self to X, else
        return self.

        Parameters
        ----------
        X : Triangle
            The triangle to be transformed

        Returns
        -------
            X_new : New triangle with transformed attributes.
        """
        X_new = X.copy()
        X_new.group_index = self._set_transform_groups(X_new)
        triangles = ["std_err_", "ldf_", "sigma_"]
        for item in triangles + ["average_", "w_", "sigma_interpolation"]:
            setattr(X_new, item, getattr(self, item))
        X_new._set_slicers()
        return X_new

    @property
    def weight_(self):
        return pd.DataFrame(
            self.w_[0, 0],
            index=self.ldf_.origin,
            columns=list(self.ldf_.development.values[:, 0]),
        )

    def _param_property(self, X, params, idx):
        from chainladder import options

        obj = X[X.origin == X.origin.min()]
        xp = X.get_array_module()
        obj.values = xp.ones(obj.shape)[..., :-1] * params[..., idx : idx + 1, :]
        obj.ddims = X.link_ratio.ddims
        obj.valuation_date = pd.to_datetime(options.ULT_VAL)
        obj.is_pattern = True
        obj.is_cumulative = False
        obj.virtual_columns.columns = {}
        obj._set_slicers()
        return obj



import chainladder as cl
from chainladder.utils.cupy import cp
import numpy as np
import copy


def test_non_vertical_line():
    true_olf = (1 - 0.5 * (184 / 365) ** 2) * 0.2
    olf_low = (
        cl.parallelogram_olf([0.20], ["7/1/2017"], grain="Y").loc["2017"].iloc[0] - 1
    )
    olf_high = (
        cl.parallelogram_olf([0.20], ["7/2/2017"], grain="Y").loc["2017"].iloc[0] - 1
    )
    assert olf_low < true_olf < olf_high


def test_vertical_line():
    olf = cl.parallelogram_olf([0.20], ["7/1/2017"], grain="Y", vertical_line=True)
    assert abs(olf.loc["2017"].iloc[0] - ((1 - 184 / 365) * 0.2 + 1)) < 0.00001


def test_triangle_json_io(clrd):
    xp = clrd.get_array_module()
    clrd2 = cl.read_json(clrd.to_json(), array_backend=clrd.array_backend)
    xp.testing.assert_array_equal(clrd.values, clrd2.values)
    xp.testing.assert_array_equal(clrd.kdims, clrd2.kdims)
    xp.testing.assert_array_equal(clrd.vdims, clrd2.vdims)
    xp.testing.assert_array_equal(clrd.odims, clrd2.odims)
    xp.testing.assert_array_equal(clrd.ddims, clrd2.ddims)
    assert np.all(clrd.valuation == clrd2.valuation)


def test_json_for_val(raa):
    x = raa.dev_to_val().to_json()
    assert cl.read_json(x) == raa.dev_to_val()


def test_estimator_json_io():
    assert (
        cl.read_json(cl.Development().to_json()).get_params()
        == cl.Development().get_params()
    )


def test_pipeline_json_io():
    pipe = cl.Pipeline(
        steps=[("dev", cl.Development()), ("model", cl.BornhuetterFerguson())]
    )
    pipe2 = cl.read_json(pipe.to_json())
    assert {item[0]: item[1].get_params() for item in pipe.get_params()["steps"]} == {
        item[0]: item[1].get_params() for item in pipe2.get_params()["steps"]
    }


def test_json_subtri(raa):
    a = cl.read_json(
        cl.Chainladder().fit_predict(raa).to_json()
    ).full_triangle_
    b = cl.Chainladder().fit_predict(raa).full_triangle_
    assert abs(a - b).max().max() < 1e-4


def test_json_df():
    x = cl.MunichAdjustment(paid_to_incurred=("paid", "incurred")).fit_transform(
        cl.load_sample("mcl")
    )
    assert abs(cl.read_json(x.to_json()).lambda_ - x.lambda_).sum() < 1e-5


def test_concat(clrd):
    tri = clrd.groupby("LOB").sum()
    assert (
        cl.concat([tri.loc["wkcomp"], tri.loc["comauto"]], axis=0)
        == tri.loc[["wkcomp", "comauto"]]
    )

def test_model_diagnostics(qtr):
    cl.model_diagnostics(cl.Chainladder().fit(qtr))

def test_concat_immutability(raa):
    u = cl.Chainladder().fit(raa).ultimate_
    l = raa.latest_diagonal
    u.columns = l.columns
    u_new = copy.deepcopy(u)
    cl.concat((l, u), axis=3)
    assert u == u_new



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import numpy as np
import pandas as pd
import warnings
from sklearn.base import BaseEstimator, TransformerMixin
from chainladder.utils import WeightedRegression
from chainladder.core.io import EstimatorIO
from chainladder.core.common import Common


class DevelopmentBase(BaseEstimator, TransformerMixin, EstimatorIO, Common):
    def _set_fit_groups(self, X):
        """Used for assigning group_index in fit"""
        backend = "numpy" if X.array_backend in ["sparse", "numpy"] else "cupy"
        if self.groupby is None:
            return X.set_backend(backend)
        if callable(self.groupby) or type(self.groupby) in [list, str, pd.Series]:
            return X.groupby(self.groupby).sum().set_backend(backend)
        else:
            raise ValueError("Cannot determine groupings.")

    def _set_transform_groups(self, X):
        """Used for assigning group_index in transform"""
        if self.groupby is None:
            return X.index
        else:
            indices = X.groupby(self.groupby).groups.indices
        return (
            pd.Series(
                {vi: k for k, v in indices.items() for vi in v},
                name=self.ldf_.key_labels[0],
            )
            .sort_index()
            .to_frame()
        )

    def _assign_n_periods_weight(self, X, n_periods):
        """Used to apply the n_periods weight"""

        def _assign_n_periods_weight_int(X, n_periods):
            xp = X.get_array_module()
            val_offset = {
                "Y": {"Y": 1},
                "Q": {"Y": 4, "Q": 1},
                "M": {"Y": 12, "Q": 3, "M": 1},
            }
            if n_periods < 1 or n_periods >= X.shape[-2] - 1:
                return X.values * 0 + 1
            else:
                val_date_min = X.valuation[X.valuation <= X.valuation_date]
                val_date_min = val_date_min.drop_duplicates().sort_values()
                z = -n_periods * val_offset[X.development_grain][X.origin_grain] - 1
                val_date_min = val_date_min[z]
                w = X[X.valuation >= val_date_min]
                return xp.nan_to_num((w / w).values) * X.nan_triangle

        xp = X.get_array_module()

        dict_map = {
            item: _assign_n_periods_weight_int(X, item) for item in set(n_periods)
        }

        conc = [
            dict_map[item][..., num : num + 1] for num, item in enumerate(n_periods)
        ]
        return xp.concatenate(tuple(conc), -1)

    def _drop_adjustment(self, X, link_ratio):
        weight = X.nan_triangle[:, :-1]

        if self.drop is not None:
            weight = weight * self._drop(X)

        if self.drop_valuation is not None:
            weight = weight * self._drop_valuation(X)

        if (self.drop_high is not None) | (self.drop_low is not None):
            n_periods_ = self._validate_axis_assumption(
                self.n_periods, X.development[:-1]
            )

            w_ = self._assign_n_periods_weight(X, n_periods_)
            w_ = w_.astype("float")
            w_[w_ == 0] = np.nan

            link_ratio_w_ = link_ratio * w_

            weight = weight * self._drop_n(
                self.drop_high,
                self.drop_low,
                X,
                link_ratio_w_,
                self.preserve,
            )

        if (self.drop_above != np.inf) | (self.drop_below != 0.00):
            weight = weight * self._drop_x(
                self.drop_above, self.drop_below, X, link_ratio, self.preserve
            )

        return weight

    # for drop_high and drop_low
    def _drop_n(self, drop_high, drop_low, X, link_ratio, preserve):
        link_ratios_len = len(link_ratio[0, 0])

        def drop_array_helper(drop_type):
            drop_type_array = np.array(link_ratios_len * [0])

            if drop_type is None:
                return np.array(link_ratios_len * [0])
            else:
                # only a single parameter is provided
                if isinstance(drop_type, int):
                    drop_type_array = np.array(link_ratios_len * [drop_type])
                elif isinstance(drop_type, bool):
                    drop_type_array = np.array(link_ratios_len * int(drop_type))

                # an array of parameters is provided
                else:
                    for index in range(len(drop_type)):
                        drop_type_array[index] = int(drop_type[index])

                # convert boolean to ints (1s)
                for index in range(len(drop_type_array)):
                    if isinstance(drop_type_array[index], bool):
                        drop_type_array[index] = int(drop_type_array[index] == True)
                    else:
                        drop_type_array[index] = drop_type_array[index]

                return drop_type_array

        drop_high_array = drop_array_helper(drop_high)
        drop_low_array = drop_array_helper(drop_low)

        link_ratio_ranks = link_ratio[0][0].argsort(axis=0).argsort(axis=0)

        weights = ~np.isnan(link_ratio[0][0].T)
        warning_flag = False

        # checking to see if the ranks are in range
        for index in range(len(drop_high_array) - 1):
            if self.n_periods == -1:
                period_guard = link_ratios_len
            else:
                period_guard = self.n_periods + 1

            max_rank = (
                min(link_ratios_len - index, period_guard) - drop_high_array[index]
            )
            min_rank = drop_low_array[index]

            index_array_weights = (link_ratio_ranks.T[index] < max_rank - 1) & (
                link_ratio_ranks.T[index] > min_rank - 1
            )

            if sum(index_array_weights) > preserve - 1:
                weights[index] = index_array_weights

            else:
                warning_flag = True

        if warning_flag:
            if preserve == 1:
                warning = (
                    "Some exclusions have been ignored. At least "
                    + str(preserve)
                    + " (use preserve = ...)"
                    + " link ratio(s) is required for development estimation."
                )
            else:
                warning = (
                    "Some exclusions have been ignored. At least "
                    + str(preserve)
                    + " link ratio(s) is required for development estimation."
                )
            warnings.warn(warning)

        return weights.T[None, None]

    # for drop_above and drop_below
    def _drop_x(self, drop_above, drop_below, X, link_ratio, preserve):
        link_ratios_len = len(link_ratio[0, 0])

        def drop_array_helper(drop_type, default_value):
            drop_type_array = np.array(link_ratios_len * [default_value])

            # only a single parameter is provided
            if isinstance(drop_type, int):
                drop_type_array = np.array(link_ratios_len * [float(drop_type)])
            elif isinstance(drop_type, float):
                drop_type_array = np.array(link_ratios_len * [drop_type])

            # an array of parameters is provided
            else:
                for index in range(len(drop_type)):
                    drop_type_array[index] = float(drop_type[index])

            return drop_type_array

        drop_above_array = drop_array_helper(drop_above, np.inf)
        drop_below_array = drop_array_helper(drop_below, 0.0)

        weights = ~np.isnan(link_ratio[0][0].T)
        warning_flag = False

        # checking to see if the ranks are in range
        for index in range(len(drop_above_array) - 1):
            max_ldf = drop_above_array[index]
            min_ldf = drop_below_array[index]

            index_array_weights = (link_ratio[0][0].T[index] <= max_ldf) & (
                link_ratio[0][0].T[index] >= min_ldf
            )

            if sum(index_array_weights) > preserve - 1:
                weights[index] = index_array_weights

            else:
                warning_flag = True

        if warning_flag:
            if preserve == 1:
                warning = (
                    "Some exclusions have been ignored. At least "
                    + str(preserve)
                    + " (use preserve = ...)"
                    + " link ratio(s) is required for development estimation."
                )
            else:
                warning = (
                    "Some exclusions have been ignored. At least "
                    + str(preserve)
                    + " link ratio(s) is required for development estimation."
                )
            warnings.warn(warning)

        return weights.T[None, None]

    def _drop_valuation(self, X):
        xp = X.get_array_module()
        if type(self.drop_valuation) is not list:
            drop_valuation = [self.drop_valuation]
        else:
            drop_valuation = self.drop_valuation
        v = pd.PeriodIndex(drop_valuation, freq=X.origin_grain).to_timestamp(how="e")
        arr = 1 - xp.nan_to_num(X[X.valuation.isin(v)].values[0, 0] * 0 + 1)
        ofill = X.shape[-2] - arr.shape[-2]
        dfill = X.shape[-1] - arr.shape[-1]
        if ofill > 0:
            arr = xp.concatenate(
                (arr, xp.repeat(xp.ones(arr.shape[-1])[None], ofill, 0)), 0
            )
        if dfill > 0:
            arr = xp.concatenate(
                (arr, xp.repeat(xp.ones(arr.shape[-2])[..., None], dfill, -1)), -1
            )
        return arr[:, :-1]

    def _drop(self, X):
        xp = X.get_array_module()
        drop = [self.drop] if type(self.drop) is not list else self.drop
        arr = X.nan_triangle.copy()
        for item in drop:
            arr[
                np.where(X.origin == item[0])[0][0],
                np.where(X.development == item[1])[0][0],
            ] = 0
        return arr[:, :-1]



import numpy as np
import pytest
import chainladder as cl

try:
    from rpy2.robjects.packages import importr
    from rpy2.robjects import r
    CL = importr("ChainLadder")
except:
    pass


def mack_r(data, alpha, est_sigma, tail):
    if tail:
        return r(
            'mack<-MackChainLadder({},alpha={}, est.sigma="{}", tail=TRUE)'.format(
                data, alpha, est_sigma
            )
        )
    else:
        return r(
            'mack<-MackChainLadder({},alpha={}, est.sigma="{}")'.format(
                data, alpha, est_sigma
            )
        )


def mack_p(data, average, est_sigma, tail):
    if tail:
        return cl.MackChainladder().fit(
            cl.TailCurve(curve="exponential").fit_transform(
                cl.Development(
                    average=average, sigma_interpolation=est_sigma
                ).fit_transform(cl.load_sample(data))
            )
        )
    else:
        return cl.MackChainladder().fit(
            cl.Development(
                average=average, sigma_interpolation=est_sigma
            ).fit_transform(cl.load_sample(data))
        )


data = ["ABC", "MW2008"]
tail = [True, False]
averages = [("simple", 0), ("volume", 1), ("regression", 2)]
est_sigma = [("log-linear", "log-linear"), ("mack", "Mack")]


def test_mack_to_triangle():
    assert (
        cl.MackChainladder()
        .fit(
            cl.TailConstant().fit_transform(
                cl.Development().fit_transform(cl.load_sample("ABC"))
            )
        )
        .summary_
        == cl.MackChainladder()
        .fit(cl.Development().fit_transform(cl.load_sample("ABC")))
        .summary_
    )

@pytest.mark.r
@pytest.mark.parametrize("data", data)
@pytest.mark.parametrize("averages", averages)
@pytest.mark.parametrize("est_sigma", est_sigma)
@pytest.mark.parametrize("tail", tail)
def test_mack_full_std_err(data, averages, est_sigma, tail, atol):
    df = mack_r(data, averages[1], est_sigma[1], tail).rx("F.se")
    p = mack_p(data, averages[0], est_sigma[0], tail).full_std_err_
    xp = p.get_array_module()
    p = p.values[0, 0, :, :][:, :-1] if not tail else p.values[0, 0, :, :]
    r = xp.array(df[0])
    assert xp.allclose(r, p, atol=atol)

@pytest.mark.r
@pytest.mark.parametrize("data", data)
@pytest.mark.parametrize("averages", averages)
@pytest.mark.parametrize("est_sigma", est_sigma)
@pytest.mark.parametrize("tail", tail)
def test_mack_process_risk(data, averages, est_sigma, tail, atol):
    df = mack_r(data, averages[1], est_sigma[1], tail).rx("Mack.ProcessRisk")
    p = mack_p(data, averages[0], est_sigma[0], tail).process_risk_
    xp = p.get_array_module()
    p = p.values[0, 0, :, :][:, :-1] if not tail else p.values[0, 0, :, :]
    r = xp.array(df[0])
    assert xp.allclose(r, p, atol=atol)

@pytest.mark.r
@pytest.mark.parametrize("data", data)
@pytest.mark.parametrize("averages", averages)
@pytest.mark.parametrize("est_sigma", est_sigma)
@pytest.mark.parametrize("tail", tail)
def test_mack_parameter_risk(data, averages, est_sigma, tail, atol):
    df = mack_r(data, averages[1], est_sigma[1], tail).rx("Mack.ParameterRisk")
    p = mack_p(data, averages[0], est_sigma[0], tail).parameter_risk_
    xp = p.get_array_module()
    p = p.values[0, 0, :, :][:, :-1] if not tail else p.values[0, 0, :, :]
    r = xp.array(df[0])
    assert xp.allclose(r, p, atol=atol)

@pytest.mark.r
@pytest.mark.parametrize("data", data)
@pytest.mark.parametrize("averages", averages)
@pytest.mark.parametrize("est_sigma", est_sigma)
@pytest.mark.parametrize("tail", tail)
def test_mack_total_process_risk(data, averages, est_sigma, tail, atol):
    df = mack_r(data, averages[1], est_sigma[1], tail).rx("Total.ProcessRisk")
    p = mack_p(data, averages[0], est_sigma[0], tail).total_process_risk_
    xp = p.get_array_module()
    p = p.values[0, 0, :, :][:, :-1] if not tail else p.values[0, 0, :, :]
    r = xp.array(df[0])[None, ...]
    assert xp.allclose(r, xp.nan_to_num(p), atol=atol)

@pytest.mark.r
@pytest.mark.parametrize("data", data)
@pytest.mark.parametrize("averages", averages)
@pytest.mark.parametrize("est_sigma", est_sigma)
@pytest.mark.parametrize("tail", tail)
def test_mack_total_parameter_risk(data, averages, est_sigma, tail, atol):
    df = mack_r(data, averages[1], est_sigma[1], tail).rx("Total.ParameterRisk")
    p = mack_p(data, averages[0], est_sigma[0], tail).total_parameter_risk_
    xp = p.get_array_module()
    p = p.values[0, 0, :, :][:, :-1] if not tail else p.values[0, 0, :, :]
    r = xp.array(df[0])[None]
    assert xp.allclose(r, xp.nan_to_num(p), atol=atol)

@pytest.mark.r
@pytest.mark.parametrize("data", data)
@pytest.mark.parametrize("averages", averages)
@pytest.mark.parametrize("est_sigma", est_sigma)
@pytest.mark.parametrize("tail", tail)
def test_mack_mack_std_err_(data, averages, est_sigma, tail, atol):
    df = mack_r(data, averages[1], est_sigma[1], tail).rx("Mack.S.E")
    p = mack_p(data, averages[0], est_sigma[0], tail).mack_std_err_
    xp = p.get_array_module()
    p = p.values[0, 0, :, :][:, :-1] if not tail else p.values[0, 0, :, :]
    r = xp.array(df[0])
    assert xp.allclose(r, xp.nan_to_num(p), atol=atol)

@pytest.mark.r
def test_mack_asymmetric():
    r("Paid <- matrix(NA, 45, 45)")
    r("Paid[seq(1,45,4),] <- qpaid")
    out = r("M <- MackChainLadder(Paid)")
    tri = cl.load_sample("quarterly")["paid"]
    xp = tri.get_array_module()
    assert round(float(xp.array(out.rx("Mack.S.E")[0])[-1, -1]), 2) == round(
        float(cl.MackChainladder().fit(tri).summary_.to_frame(origin_as_datetime=False).iloc[-1, -1]), 2
    )

def test_mack_malformed():
    a  = cl.load_sample('raa')
    b = a.iloc[:, :, :-1]
    x = cl.MackChainladder().fit(a) 
    y = cl.MackChainladder().fit(b)
    assert x.process_risk_.iloc[:,:,:-1] == y.process_risk_


import numpy as np
import pytest
import chainladder as cl

try:
    from rpy2.robjects.packages import importr
    from rpy2.robjects import r
    CL = importr("ChainLadder")
except:
    pass


def mack_r(data, alpha, est_sigma):
    return r(
        'mack<-MackChainLadder({},alpha={}, est.sigma="{}", tail=TRUE)'.format(
            data, alpha, est_sigma
        )
    )


def mack_p(data, average, est_sigma):
    return cl.TailCurve(curve="exponential").fit_transform(
        cl.Development(average=average, sigma_interpolation=est_sigma).fit_transform(
            cl.load_sample(data)
        )
    )


def mack_p_no_tail(data, average, est_sigma):
    return cl.Development(average=average, sigma_interpolation=est_sigma).fit_transform(
        cl.load_sample(data)
    )


data = ["RAA", "ABC", "GenIns", "MW2008", "MW2014"]
# M3IR5 in R fails silently on exponential tail. Python actually computes it.
averages = [("simple", 0), ("volume", 1), ("regression", 2)]
est_sigma = [("mack", "Mack"), ("log-linear", "log-linear")]


@pytest.mark.r
@pytest.mark.parametrize("data", data)
@pytest.mark.parametrize("averages", averages)
@pytest.mark.parametrize("est_sigma", est_sigma)
def test_mack_tail_ldf(data, averages, est_sigma, atol):
    p = mack_p(data, averages[0], est_sigma[0]).ldf_.set_backend("numpy", inplace=True)
    xp = p.get_array_module()
    r = xp.array(mack_r(data, averages[1], est_sigma[1]).rx("f"))
    p = xp.concatenate(
        (
            p.values[0, 0, :, :][:, :-2],
            xp.prod(p.values[0, 0, :, :][:, -2:], -1, keepdims=True),
        ),
        -1,
    )
    p = xp.unique(p, axis=-2)
    assert xp.allclose(r, p, atol=atol)


@pytest.mark.r
@pytest.mark.parametrize("data", data)
@pytest.mark.parametrize("averages", averages)
@pytest.mark.parametrize("est_sigma", est_sigma)
def test_mack_tail_sigma(data, averages, est_sigma, atol):
    p = mack_p(data, averages[0], est_sigma[0]).sigma_
    xp = p.get_array_module()
    r = xp.array(mack_r(data, averages[1], est_sigma[1]).rx("sigma"))
    p = xp.unique(p.values[0, 0, :, :], axis=-2)
    assert xp.allclose(r, p, atol=atol)


@pytest.mark.r
@pytest.mark.parametrize("data", data)
@pytest.mark.parametrize("averages", averages)
@pytest.mark.parametrize("est_sigma", est_sigma)
def test_mack_tail_std_err(data, averages, est_sigma, atol):
    p = mack_p(data, averages[0], est_sigma[0]).std_err_
    xp = p.get_array_module()
    r = xp.array(mack_r(data, averages[1], est_sigma[1]).rx("f.se"))
    p = xp.unique(p.values[0, 0, :, :], axis=-2)
    assert xp.allclose(r, p, atol=atol)


@pytest.mark.r
@pytest.mark.parametrize("data", data)
@pytest.mark.parametrize("averages", averages[0:1])
@pytest.mark.parametrize("est_sigma", est_sigma[0:1])
def test_tail_doesnt_mutate_std_err(data, averages, est_sigma):
    p = mack_p(data, averages[0], est_sigma[0]).std_err_
    xp = p.get_array_module()
    p_no_tail = mack_p_no_tail(data, averages[0], est_sigma[0]).std_err_.values
    xp.testing.assert_array_equal(p_no_tail, p.values[:, :, :, :-1])


@pytest.mark.r
@pytest.mark.parametrize("data", data)
@pytest.mark.parametrize("averages", averages[0:1])
@pytest.mark.parametrize("est_sigma", est_sigma[0:1])
def test_tail_doesnt_mutate_ldf_(data, averages, est_sigma):
    p = mack_p(data, averages[0], est_sigma[0]).ldf_
    xp = p.get_array_module()
    p_no_tail = mack_p_no_tail(data, averages[0], est_sigma[0]).ldf_.values
    xp.testing.assert_array_equal(
        p_no_tail, p.values[..., : len(cl.load_sample(data).ddims) - 1]
    )


@pytest.mark.r
@pytest.mark.parametrize("data", data)
@pytest.mark.parametrize("averages", averages[0:1])
@pytest.mark.parametrize("est_sigma", est_sigma[0:1])
def test_tail_doesnt_mutate_sigma_(data, averages, est_sigma):
    p = mack_p(data, averages[0], est_sigma[0]).sigma_
    xp = p.get_array_module()
    p_no_tail = mack_p_no_tail(data, averages[0], est_sigma[0]).sigma_.values
    xp.testing.assert_array_equal(p_no_tail, p.values[:, :, :, :-1])


def test_fit_period():
    tri = cl.load_sample("tail_sample")
    dev = cl.Development(average="simple").fit_transform(tri)
    assert (
        round(
            cl.TailCurve(fit_period=(tri.ddims[-7], None), extrap_periods=10)
            .fit(dev)
            .cdf_["paid"]
            .set_backend("numpy", inplace=True)
            .values[0, 0, 0, -2],
            3,
        )
        == 1.044
    )



import numpy as np
import chainladder as cl
import pytest


@pytest.fixture
def triangle_data():
    clrd = cl.load_sample("clrd")[["CumPaidLoss", "EarnedPremDIR"]]
    clrd = clrd[clrd["LOB"] == "wkcomp"]
    return clrd


@pytest.fixture
def estimators():
    bcl = cl.Chainladder()
    bf = cl.BornhuetterFerguson()
    cc = cl.CapeCod()

    estimators = [('bcl', bcl), ('bf', bf), ('cc', cc)]

    return estimators


array_weight = np.array([[1, 2, 3]] * 4 + [[0, 0.5, 0.5]] * 3 + [[0, 0, 1]] * 3)

list_weight = [[[1, 2, 3]] * 4 + [[0, 0.5, 0.5]] * 3 + [[0, 0, 1]] * 3]

callable_weight = lambda origin: np.where(
        origin.year < 1992,
        (1, 2, 3),
        np.where(origin.year > 1994, (0, 0, 1), (0, 0.5, 0.5))
    )

dict_weight = {
    '1992': (0, 0.5, 0.5),
    '1993': (0, 0.5, 0.5),
    '1994': (0, 0.5, 0.5),
    '1995': (0, 0, 1),
    '1996': (0, 0, 1),
    '1997': (0, 0, 1),
    }


@pytest.fixture(params=[array_weight, list_weight, callable_weight, dict_weight])
def weights(request):
    return request.param


def test_voting_ultimate(triangle_data, estimators, weights):
    bcl_ult = cl.Chainladder().fit(
        triangle_data["CumPaidLoss"].sum(),
        ).ultimate_
    bf_ult = cl.BornhuetterFerguson().fit(
        triangle_data["CumPaidLoss"].sum(),
        sample_weight=triangle_data["EarnedPremDIR"].sum().latest_diagonal
        ).ultimate_
    cc_ult = cl.CapeCod().fit(
        triangle_data["CumPaidLoss"].sum(),
        sample_weight=triangle_data["EarnedPremDIR"].sum().latest_diagonal
        ).ultimate_

    vot_ult = cl.VotingChainladder(
        estimators=estimators,
        weights=weights,
        default_weighting=(1, 2, 3)
    ).fit(
        triangle_data["CumPaidLoss"].sum(),
        sample_weight=triangle_data["EarnedPremDIR"].sum().latest_diagonal,
    ).ultimate_

    direct_weight = np.array([[1, 2, 3]] * 4 + [[0, 0.5, 0.5]] * 3 + [[0, 0, 1]] * 3)
    direct_weight = direct_weight[..., np.newaxis]

    assert abs(
        (
            (
                bcl_ult * direct_weight[..., 0, :] +
                bf_ult * direct_weight[..., 1, :] +
                cc_ult * direct_weight[..., 2, :]
            ) / direct_weight.sum(axis=-2)
        ).sum() - vot_ult.sum()
    ) < 1


def test_different_backends(triangle_data, estimators, weights):
    model = cl.VotingChainladder(
        estimators=estimators,
        weights=weights,
        default_weighting=(1, 2, 3)
    ).fit(
        triangle_data["CumPaidLoss"].sum().set_backend("numpy"),
        sample_weight=triangle_data["EarnedPremDIR"].sum().latest_diagonal.set_backend("numpy"),
    )
    assert (
        abs(
            (
                model.predict(
                    triangle_data["CumPaidLoss"].sum().set_backend("sparse"),
                    sample_weight=triangle_data["EarnedPremDIR"].sum().latest_diagonal.set_backend(
                        "sparse"
                    )
                ).ultimate_.sum()
                - model.ultimate_.sum()
            )
        )
        < 1
    )


def test_weight_broadcasting(triangle_data, estimators, weights):
    mid_dim_weights = np.array([[[1, 2, 3]] * 4 + [[0, 0.5, 0.5]] * 3 + [[0, 0, 1]] * 3] * 1)
    max_dim_weights = np.array(mid_dim_weights * 132)

    min_dim_ult = cl.VotingChainladder(estimators=estimators, weights=weights).fit(
        triangle_data['CumPaidLoss'],
        sample_weight=triangle_data["EarnedPremDIR"].latest_diagonal,
    ).ultimate_.sum()
    mid_dim_ult = cl.VotingChainladder(estimators=estimators, weights=mid_dim_weights).fit(
        triangle_data['CumPaidLoss'],
        sample_weight=triangle_data["EarnedPremDIR"].latest_diagonal,
    ).ultimate_.sum()
    max_dim_ult = cl.VotingChainladder(estimators=estimators, weights=max_dim_weights).fit(
        triangle_data['CumPaidLoss'],
        sample_weight=triangle_data["EarnedPremDIR"].latest_diagonal,
    ).ultimate_.sum()
    assert (abs(min_dim_ult - mid_dim_ult - max_dim_ult) < 1)



import chainladder as cl


def test_struhuss():
    X = cl.load_sample("cc_sample")["loss"]
    X = cl.TailConstant(tail=1 / 0.85).fit_transform(cl.Development().fit_transform(X))
    sample_weight = cl.load_sample("cc_sample")["exposure"].latest_diagonal
    ibnr = int(
        cl.CapeCod(trend=0.07, decay=0.75)
        .fit(X, sample_weight=sample_weight)
        .ibnr_.sum()
    )
    assert ibnr == 17052


def test_groupby(clrd):
    clrd = clrd[clrd['LOB']=='comauto']
    # But only the top 10 get their own CapeCod aprioris. Smaller companies get grouped together
    top_10 = clrd['EarnedPremDIR'].groupby('GRNAME').sum().latest_diagonal
    top_10 = top_10.loc[..., '1997', :].to_frame(origin_as_datetime=True).nlargest(10)
    cc_groupby = clrd.index['GRNAME'].map(lambda x: x if x in top_10.index else 'Remainder')
    idx = clrd.index
    idx['Top 10'] = cc_groupby
    clrd.index = idx

    # All companies share the same development factors regardless of size
    X = cl.Development().fit(clrd['CumPaidLoss'].sum()).transform(clrd['CumPaidLoss'])
    sample_weight=clrd['EarnedPremDIR'].latest_diagonal
    a = cl.CapeCod(groupby='Top 10', decay=0.98, trend=0.02).fit(X, sample_weight=sample_weight).ibnr_.groupby('Top 10').sum().sort_index()
    b = cl.CapeCod(decay=0.98, trend=0.02).fit(X.groupby('Top 10').sum(), sample_weight=sample_weight.groupby('Top 10').sum()).ibnr_.sort_index()
    xp = a.get_array_module()
    b = b.set_backend(a.array_backend)
    xp.allclose(xp.nan_to_num(a.values), xp.nan_to_num(b.values), atol=1e-5)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from sklearn.base import BaseEstimator, TransformerMixin
from chainladder.core.io import EstimatorIO
import pandas as pd


class Trend(BaseEstimator, TransformerMixin, EstimatorIO):
    """
    Estimator to create and apply trend factors to a Triangle object.  Allows
    for compound trends as well as storage of the trend matrix to be used in
    other estimators, such as `CapeCod`.

    Parameters
    ----------

    trends: list-like
        The list containing the annual trends expressed as a decimal. For example,
        5% decrease should be stated as -0.05
    dates: list of date-likes
        A list-like of (start, end) dates to correspond to the `trend` list.
    axis: str (options: [‘origin’, ‘valuation’])
        The axis on which to apply the trend

    Attributes
    ----------

    trend_:
        A triangle representation of the trend factors

    """

    def __init__(self, trends=0.0, dates=None, axis="origin"):
        self.trends = trends
        self.dates = dates
        self.axis = axis

    def fit(self, X, y=None, sample_weight=None):
        """Fit the model with X.

        Parameters
        ----------
        X: Triangle-like
            Data to which the model will be applied.
        y: Ignored
        sample_weight: Ignored

        Returns
        -------
        self: object
            Returns the instance itself.
        """
        trends = self.trends if type(self.trends) is list else [self.trends]
        dates = [(None, None)] if self.dates is None else self.dates
        dates = [dates] if type(dates) is not list else dates
        if type(dates[0]) is not tuple:
            raise AttributeError(
                'Dates must be specified as a tuple of start and end dates')
        self.trend_ = X.copy()
        for i, trend in enumerate(trends):
            self.trend_ = self.trend_.trend(
                trend, self.axis,
                start=dates[i][0], end=dates[i][1])
        self.trend_ = self.trend_ / X
        return self

    def transform(self, X, y=None, sample_weight=None):
        """ If X and self are of different shapes, align self to X, else
        return self.

        Parameters
        ----------
        X: Triangle
            The triangle to be transformed

        Returns
        -------
            X_new: New triangle with transformed attributes.
        """
        X_new = X.copy()
        triangles = ["trend_"]
        for item in triangles:
            setattr(X_new, item, getattr(self, item))
        X_new._set_slicers()
        return X_new



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import numpy as np
from chainladder.utils import WeightedRegression
from chainladder.development import DevelopmentBase, Development


class TailBase(DevelopmentBase):
    """ Base class for all tail methods.  Tail objects are equivalent
        to development objects with an additional set of tail statistics"""

    def fit(self, X, y=None, sample_weight=None):
        obj = X.copy()
        if "ldf_" not in obj:
            obj = Development().fit_transform(obj)
        xp = obj.ldf_.get_array_module()
        m = int(self.projection_period / 12)
        self._ave_period = {"Y": (1 * m, 12), "Q": (4 * m, 3), "M": (12 * m, 1)}[
            obj.development_grain
        ]
        t_ddims = [
            (item + 1) * self._ave_period[1] + obj.ldf_.ddims[-1]
            for item in range(self._ave_period[0]+1)
        ]
        ddims = np.concatenate((obj.ldf_.ddims, t_ddims), 0,)
        self.ldf_ = obj.ldf_.copy()
        tail = xp.ones(self.ldf_.shape)[..., -1:]
        tail = xp.repeat(tail, self._ave_period[0] + 1, -1)
        self.ldf_.values = xp.concatenate((self.ldf_.values, tail), -1)
        self.ldf_.ddims = ddims
        if hasattr(obj, "sigma_"):
            zeros = (obj.sigma_.iloc[..., -1:] * 0).values
            self.sigma_ = getattr(obj, "sigma_").copy()
            self.sigma_.values = xp.concatenate((self.sigma_.values, zeros), -1)
            self.std_err_ = getattr(obj, "std_err_").copy()
            self.std_err_.values = xp.concatenate((self.std_err_.values, zeros), -1)
            self.sigma_.ddims = self.std_err_.ddims = self.ldf_.ddims[:obj.shape[2]]
            self.sigma_._set_slicers()
            self.std_err_._set_slicers()
        if hasattr(obj, "average_"):
            self.average_ = obj.average_
        else:
            self.average_ = None
        self.ldf_._set_slicers()
        return self

    def transform(self, X):
        """ If X and self are of different shapes, align self to X, else
        return self.

        Parameters
        ----------
        X : Triangle
            The triangle to be transformed

        Returns
        -------
            X_new : New triangle with transformed attributes.
        """
        X_new = X.copy()
        triangles = ["ldf_", "std_err_", "sigma_"]
        for item in triangles + ["tail_", "_ave_period", "average_"]:
            if hasattr(self, item):
                setattr(X_new, item, getattr(self, item))
        X_new._set_slicers()
        return X_new

    def _get_tail_prediction(self, tail_ldf):
        xp = self.ldf_.get_array_module()
        accum_point = self.ldf_.shape[-1] - 1
        ave = 1 + tail_ldf[..., :accum_point]
        all = xp.prod(1 + tail_ldf[..., accum_point:], -1)[..., None]
        tail = xp.concatenate((ave, all), -1)
        return tail

    def _get_initial_ldf(self, xp, tail):
        """ Quadratic series expansion solution to return seed LDF for tail"""
        arr = self.decay ** xp.arange(1000)
        a = xp.sum(arr ** 2)
        b = xp.sum(arr)
        c = -xp.log(tail)
        return (-b + xp.sqrt(b ** 2 - 4 * a * c)) / (2 * a)

    def _apply_decay(self, X, tail, attach_idx=None):
        """ Created Tail vector with decay over time. """
        xp = self.ldf_.get_array_module()
        if attach_idx:
            decay_range = self.ldf_.shape[-1] - attach_idx
        else:
            decay_range = self.ldf_.shape[-1] - X.shape[-1] + 1
        if xp.max(xp.array(tail)) == 1.0:
            ldfs = 1 + 0 * (self.decay ** xp.arange(1000))
        else:
            ldfs = 1 + self._get_initial_ldf(xp, tail) * (self.decay ** xp.arange(1000))
        ldfs = ldfs[..., :decay_range]
        tail = tail / xp.prod(ldfs[..., :-1], axis=-1, keepdims=True)
        ldfs = xp.concatenate((ldfs[..., :-1], tail), axis=-1)
        self.ldf_.values = xp.concatenate(
            (
                self.ldf_.values[..., :-decay_range],
                (xp.nan_to_num(self.ldf_.values[..., -decay_range:]) * 0 + 1) * ldfs,
            ),
            axis=-1,
        )
        return self

    def _get_tail_stats(self, X):
        """ Method to approximate the tail sigma using
        log-linear extrapolation applied to tail average period
        """
        from chainladder.utils.utility_functions import num_to_nan
        if not hasattr(X, 'sigma_'):
            self.sigma_ = None
            self.std_err_ = None
        else:
            time_pd = self._get_tail_weighted_time_period(X)
            xp = X.sigma_.get_array_module()
            reg = WeightedRegression(axis=3, xp=xp).fit(None, xp.log(X.sigma_.values), None)
            sigma_ = xp.exp(time_pd * reg.slope_ + reg.intercept_)
            y = X.std_err_.values
            y = num_to_nan(y)
            reg = WeightedRegression(axis=3, xp=xp).fit(None, xp.log(y), None)
            std_err_ = xp.exp(time_pd * reg.slope_ + reg.intercept_)
            if self.tail_.values.flatten().sum() / xp.prod(self.tail_.shape) == 1.0:
                # If no tail, assume no variation
                sigma_ = sigma_ * 0
                std_err_ = std_err_* 0
            self.sigma_.values = xp.concatenate(
                (self.sigma_.values[..., :-1], sigma_[..., -1:]), axis=-1
            )
            self.std_err_.values = xp.concatenate(
                (self.std_err_.values[..., :-1], std_err_[..., -1:]), axis=-1
            )

    def _get_tail_weighted_time_period(self, X):
        """ Method to approximate the weighted-average development age of tail
        using log-linear extrapolation

        Returns: float32
        """
        y = X.ldf_.values.copy()
        xp = X.ldf_.get_array_module()
        y[y <= 1] = xp.nan
        reg = WeightedRegression(axis=3, xp=xp).fit(None, xp.log(y - 1), None)
        tail = xp.prod(
            self.ldf_.values[..., -self._ave_period[0] - 1 :], -1, keepdims=True
        )
        reg = WeightedRegression(axis=3, xp=xp).fit(None, xp.log(y - 1), None)
        tail = tail if tail.max() > 1 else 1.001
        time_pd = (xp.log(tail - 1) - reg.intercept_) / reg.slope_
        return time_pd

    @staticmethod
    def _tail_(self):
        df = self.cdf_[
            self.cdf_.development
            == self.cdf_.development.iloc[-1 - self._ave_period[0]]
        ]
        if np.all(df.values.min(axis=2) == df.values.max(axis=2)):
            df = df.iloc[..., 0, :].to_frame(origin_as_datetime = False)
        return df

    @property
    def tail_(self):
        return TailBase._tail_(self)




import numpy as np
import pytest


def test_slice_by_boolean(clrd):
    assert (
        clrd[clrd["LOB"] == "ppauto"].loc["Wolverine Mut Ins Co"]["CumPaidLoss"]
        == clrd.loc["Wolverine Mut Ins Co"].loc["ppauto"]["CumPaidLoss"])


def test_slice_by_loc(clrd):
    assert clrd.loc["Aegis Grp"].loc["comauto"].index.iloc[0, 0] == "comauto"


def test_slice_origin(raa):
    assert raa[raa.origin > "1985"].shape == (1, 1, 5, 10)
    raa.loc[..., raa.origin<='1994', :]


def test_slice_development(raa):
    assert raa[raa.development < 72].shape == (1, 1, 10, 5)
    raa.loc[..., 24:]


def test_slice_by_loc_iloc(clrd):
    assert clrd.groupby("LOB").sum().loc["comauto"].index.iloc[0, 0] == "comauto"
    assert len(clrd.loc[clrd.index.iloc[150]]) == 1


def test_slicers_honor_order(clrd):
    clrd = clrd.groupby("LOB").sum()
    assert clrd.iloc[[1, 0], :].iloc[0, 1] == clrd.iloc[1, 1]  # row
    assert clrd.iloc[[1, 0], [1, 0]].iloc[0, 0] == clrd.iloc[1, 1]  # col
    assert clrd.loc[:, ["CumPaidLoss", "IncurLoss"]].iloc[0, 0] == clrd.iloc[0, 1]
    assert (
        clrd.loc[["ppauto", "medmal"], ["CumPaidLoss", "IncurLoss"]].iloc[0, 0]
        == clrd.iloc[3]["CumPaidLoss"]
    )
    assert (
        clrd.loc[clrd["LOB"] == "comauto", ["CumPaidLoss", "IncurLoss"]]
        == clrd[clrd["LOB"] == "comauto"].iloc[:, [1, 0]]
    )
    assert clrd.groupby("LOB").sum() == clrd


def test_backends(clrd):
    clrd = clrd[["CumPaidLoss", "EarnedPremDIR"]]
    a = clrd.iloc[1, 0].set_backend("sparse").dropna()
    b = clrd.iloc[1, 0].dropna()
    assert a == b


def test_union_columns(clrd):
    assert clrd.iloc[:, :3] + clrd.iloc[:, 3:] == clrd


def test_4loc(clrd):
    clrd = clrd.groupby("LOB").sum()
    assert (
        clrd.iloc[:3, :2, 0, 0]
        == clrd[clrd.origin == clrd.origin.min()][
            clrd.development == clrd.development.min()
        ].loc["comauto":"othliab", :"CumPaidLoss", :, :]
    )
    assert (
        clrd.iloc[:3, :2, 0:1, -1]
        == clrd[clrd.development == clrd.development.max()].loc[
            "comauto":"othliab", :"CumPaidLoss", "1988", :
        ]
    )


def test_loc_ellipsis(clrd):
    assert (
        clrd.loc["Aegis Grp"] == clrd.loc["Adriatic Ins Co":"Aegis Grp"].loc["Aegis Grp"]
    )
    assert clrd.loc["Aegis Grp", ..., :] == clrd.loc["Aegis Grp"]
    assert clrd.loc[..., 24:] == clrd.loc[..., :, 24:]
    assert clrd.loc[:, ..., 24:] == clrd.loc[..., :, 24:]
    assert clrd.loc[:, "CumPaidLoss"] == clrd.loc[:, "CumPaidLoss", ...]
    assert clrd.loc[..., "CumPaidLoss", :, :] == clrd.loc[:, "CumPaidLoss", :, :]


def test_missing_first_lag(raa):
    x = raa.copy()
    x.values[:, :, :, 0] = 0
    x = x.sum(0)
    x.link_ratio.shape == (1, 1, 9, 9)


def test_reverse_slice_integrity(clrd):
    assert clrd.iloc[::-1, ::-1].shape == clrd.shape
    assert np.all(clrd.iloc[:, ::-1].columns.values == clrd.columns[::-1])
    assert clrd.iloc[clrd.index.index[::-1]] + clrd == 2 * clrd


def test_loc_tuple(clrd):
    assert len(clrd.loc[("Adriatic Ins Co", "othliab")]) == 1
    assert clrd.loc[clrd.index] == clrd


def test_at_iat(raa):
    raa1 = raa.copy()
    raa2 = raa.copy()
    raa1.at['Total','values', '1985', 120]
    raa1.at['Total','values', '1985', 120] = 5
    raa1.at['Total','values', '1985', 12] = 5
    raa2.iat[0, 0, 4, -1] = 5
    raa2.iat[0, 0, 4, -1]
    raa2.iat[-1, -1, 4, 0] = 5
    assert raa1 == raa2


@pytest.mark.xfail
def test_sparse_at_iat(prism):
    prism.iloc[0, 0, 0, 0] = 1.0


def test_sparse_at_iat1(prism):
    t = prism.copy()
    t.iat[0, 0, 0, 0] = 5.0
    t.at[12138, 'reportedCount', '2008-01', 1] = 0
    assert t == prism


def test_sparse_column_assignment(prism):
    t = prism.copy()
    out = t['Paid']
    t['Paid2'] = t['Paid'] # New from physical
    t['Paid2'] = lambda x: x['Paid'] # Existing from virtual
    t['Paid'] = t['Paid2'] # Existing from physical
    t['Paid3'] = lambda t: t['Paid'] # New from virtual
    assert out == t['Paid']
    assert t.shape == (34244, 6, 120, 120)



import pytest
import chainladder as cl


def test_heatmap_render(raa):
    """ The heatmap method should render correctly given the sample."""
    return raa.heatmap()


def test_to_frame(raa):
    try:
        cl.Chainladder().fit(raa).cdf_.to_frame()
        cl.Chainladder().fit(raa).cdf_.to_frame(origin_as_datetime=False)
        cl.Chainladder().fit(raa).cdf_.to_frame(origin_as_datetime=True)
        cl.Chainladder().fit(raa).ultimate_.to_frame()
        cl.Chainladder().fit(raa).ultimate_.to_frame(origin_as_datetime=False)
        cl.Chainladder().fit(raa).ultimate_.to_frame(origin_as_datetime=True)

    except:
        assert False



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from chainladder.development.base import DevelopmentBase
import pandas as pd


class DevelopmentConstant(DevelopmentBase):
    """ A Estimator that allows for including of external patterns into a
        Development style model. When this estimator is fit against a triangle,
        only the grain of the existing triangle is retained.

    Parameters
    ----------
    patterns: dict or callable
        A dictionary key:value representation of age(in months):value. If callable
        is supplied, callable must return a dict for each element of the callable axis
    style: string, optional (default='ldf')
        Type of pattern given to the Estimator. Options include 'cdf' or 'ldf'.
    callable_axis: 0 or 1
        If a callable is supplied, the axis (index or column) along which to apply
        the callable. If patterns is not a callable, then this parameter is ignored.
    groupby:
        option to group levels of the triangle index together for the purposes
        estimating patterns.  If omitted, each level of the triangle
        index will receive its own patterns.

    Attributes
    ----------
    ldf_: Triangle
        The estimated loss development patterns
    cdf_: Triangle
        The estimated cumulative development patterns
    """

    def __init__(self, patterns=None, style="ldf", callable_axis=0, groupby=None):
        self.patterns = patterns
        self.style = style
        self.callable_axis = callable_axis
        self.groupby = groupby

    def fit(self, X, y=None, sample_weight=None):
        """Fit the model with X.
        Parameters
        ----------
        X : Triangle-like
            Set of LDFs to which the munich adjustment will be applied.
        y : Ignored
        sample_weight : Ignored
        Returns
        -------
        self : object
            Returns the instance itself.
        """
        from chainladder import options
        if X.is_cumulative == False:
            obj = self._set_fit_groups(X).incr_to_cum().val_to_dev().copy()
        else:
            obj = self._set_fit_groups(X).val_to_dev().copy()
        xp = obj.get_array_module()
        obj = obj.iloc[..., :1, :-1]*0+1
        if callable(self.patterns):
            ldf = obj.index.apply(self.patterns, axis=self.callable_axis)
            ldf = (
                pd.concat(ldf.apply(pd.DataFrame, index=[0]).values, axis=0)
                  .fillna(1)[obj.ddims].values)
            ldf = xp.array(ldf[:, None, None, :])
        else:
            ldf = xp.array([float(self.patterns[item]) for item in obj.ddims])
            ldf = ldf[None, None, None, :]
        if self.style == "cdf":
            ldf = xp.concatenate((ldf[..., :-1] / ldf[..., 1:], ldf[..., -1:]), -1)
        obj = obj * ldf
        obj._set_slicers()
        self.ldf_ = obj
        self.ldf_.is_pattern = True
        self.ldf_.is_cumulative = False
        self.ldf_.valuation_date = pd.to_datetime(options.ULT_VAL)
        return self

    def transform(self, X):
        """ If X and self are of different shapes, align self to X, else
        return self.

        Parameters
        ----------
        X : Triangle
            The triangle to be transformed

        Returns
        -------
            X_new : New triangle with transformed attributes.
        """
        X_new = X.copy()
        X_new.group_index = self._set_transform_groups(X_new)
        triangles = ["ldf_"]
        for item in triangles:
            setattr(X_new, item, getattr(self, item))
        X_new._set_slicers()
        return X_new



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import pandas as pd
import numpy as np
import warnings
import re

try:
    from IPython.core.display import HTML
except:
    HTML = None


class TriangleDisplay:
    def __repr__(self):
        if (self.values.shape[0], self.values.shape[1]) == (1, 1):
            data = self._repr_format()
            return data.to_string()
        else:
            return self._summary_frame().__repr__()

    def _summary_frame(self):
        return pd.Series(
            [
                self.valuation_date.strftime("%Y-%m"),
                "O" + self.origin_grain + "D" + self.development_grain,
                self.shape,
                self.key_labels,
                list(self.vdims),
            ],
            index=["Valuation:", "Grain:", "Shape:", "Index:", "Columns:"],
            name="Triangle Summary",
        ).to_frame()

    def _repr_html_(self):
        """ Jupyter/Ipython HTML representation """
        if (self.values.shape[0], self.values.shape[1]) == (1, 1):
            data = self._repr_format()
            fmt_str = self._get_format_str(data)
            default = (
                data.to_html(
                    max_rows=pd.options.display.max_rows,
                    max_cols=pd.options.display.max_columns,
                    float_format=fmt_str.format,
                )
                .replace("nan", "")
                .replace("NaN", "")
            )
            return default
        else:
            return self._summary_frame().to_html(
                max_rows=pd.options.display.max_rows,
                max_cols=pd.options.display.max_columns,
            )

    def _get_format_str(self, data):
        if np.all(np.isnan(data)):
            return ""
        elif np.nanmean(abs(data)) < 10:
            return "{0:,.4f}"
        elif np.nanmean(abs(data)) < 1000:
            return "{0:,.2f}"
        else:
            return "{:,.0f}"

    def _repr_format(self, origin_as_datetime=False):
        out = self.compute().set_backend("numpy").values[0, 0]
        if origin_as_datetime:
            origin = pd.to_datetime(self.origin.copy(), errors='ignore')
        else:
            origin = self.origin.copy()
        origin.name = None

        if self.origin_grain == "S":
            origin_formatted = [""] * len(origin)
            for origin_index in range(len(origin)):
                origin_formatted[origin_index] = (
                    origin.astype("str")[origin_index]
                    .replace("Q1", "H1")
                    .replace("Q3", "H2")
                )
            origin = origin_formatted

        development = self.development.copy()
        development.name = None
        return pd.DataFrame(out, index=origin, columns=development)

    def heatmap(self, cmap="coolwarm", low=0, high=0, axis=0, subset=None):
        """ Color the background in a gradient according to the data in each
        column (optionally row). Requires matplotlib

        Parameters
        ----------

        cmap : str or colormap
            matplotlib colormap
        low, high : float
            compress the range by these values.
        axis : int or str
            The axis along which to apply heatmap
        subset : IndexSlice
            a valid slice for data to limit the style application to

        Returns
        -------
            Ipython.display.HTML

        """
        if (self.values.shape[0], self.values.shape[1]) == (1, 1):
            data = self._repr_format()
            fmt_str = self._get_format_str(data)

            axis = self._get_axis(axis)

            raw_rank = data.rank(axis=axis)
            shape_size = data.shape[axis]
            rank_size = data.rank(axis=axis).max(axis=axis)
            gmap = (raw_rank - 1).div(rank_size - 1, axis=not axis) * (
                shape_size - 1
            ) + 1
            gmap = gmap.replace(np.nan, (shape_size + 1) / 2)
            if pd.__version__ >= "1.3":
                default_output = (
                    data.style.format(fmt_str)
                    .background_gradient(
                        cmap=cmap,
                        low=low,
                        high=high,
                        axis=None,
                        subset=subset,
                        gmap=gmap,
                    )
                    .to_html()
                )
            else:
                default_output = (
                    data.style.format(fmt_str)
                    .background_gradient(cmap=cmap, low=low, high=high, axis=axis,)
                    .render()
                )
            output_xnan = re.sub("<td.*nan.*td>", "<td></td>", default_output)
        else:
            raise ValueError("heatmap only works with single triangles")
        if HTML:
            return HTML(output_xnan)
        elif HTML is None:
            raise ImportError("heatmap requires IPython")



import chainladder as cl

raa = cl.load_sample("RAA")
raa_1989 = raa[raa.valuation < raa.valuation_date]
cl_ult = cl.Chainladder().fit(raa).ultimate_  # Chainladder Ultimate
apriori = cl_ult * 0 + (float(cl_ult.sum()) / 10)  # Mean Chainladder Ultimate
apriori_1989 = apriori[apriori.origin < "1990"]


def test_cc_predict():
    cc = cl.CapeCod().fit(raa_1989, sample_weight=apriori_1989)
    cc.predict(raa, sample_weight=apriori)


def test_bf_predict():
    cc = cl.BornhuetterFerguson().fit(raa_1989, sample_weight=apriori_1989)
    cc.predict(raa, sample_weight=apriori)


def test_mack_predict():
    mack = cl.MackChainladder().fit(raa_1989)
    mack.predict(raa_1989)
    # mack.predict(raa)


def test_bs_random_state_predict(clrd):
    tri = (
        clrd
        .groupby("LOB")
        .sum()
        .loc["wkcomp", ["CumPaidLoss", "EarnedPremNet"]]
    )
    X = cl.BootstrapODPSample(random_state=100).fit_transform(tri["CumPaidLoss"])
    bf = cl.BornhuetterFerguson(apriori=0.6, apriori_sigma=0.1, random_state=42).fit(
        X, sample_weight=tri["EarnedPremNet"].latest_diagonal
    )
    assert (
        abs(
            bf.predict(X, sample_weight=tri["EarnedPremNet"].latest_diagonal)
            .ibnr_.sum()
            .sum()
            / bf.ibnr_.sum().sum()
            - 1
        )
        < 5e-3
    )


def test_basic_transform(raa):
    cl.Development().fit_transform(raa)
    cl.ClarkLDF().fit_transform(raa)
    cl.TailClark().fit_transform(raa)
    cl.TailBondy().fit_transform(raa)
    cl.TailConstant().fit_transform(raa)
    cl.TailCurve().fit_transform(raa)
    cl.BootstrapODPSample().fit_transform(raa)
    cl.IncrementalAdditive().fit_transform(raa, sample_weight=raa.latest_diagonal)

def test_misaligned_index(prism):
    prism = prism['Paid']
    model = cl.Chainladder().fit(cl.Development(groupby=['Line', 'Type']).fit_transform(prism))
    a = model.ultimate_.loc[prism.index.iloc[:10]].sum().sum()
    b = model.predict(prism.iloc[:10]).ultimate_.sum().sum()
    assert abs(a - b) < 1e-5


def test_misaligned_index2(clrd):
    clrd = clrd['CumPaidLoss']
    w = cl.load_sample('clrd')['EarnedPremDIR'].latest_diagonal
    bcl = cl.Chainladder().fit(cl.Development(groupby=['LOB']).fit_transform(clrd))
    bbk = cl.Benktander().fit(cl.Development(groupby=['LOB']).fit_transform(clrd), sample_weight=w)
    bcc = cl.CapeCod().fit(cl.Development(groupby=['LOB']).fit_transform(clrd), sample_weight=w)

    a = bcl.ultimate_.iloc[:10].sum().sum()
    b = bcl.predict(clrd.iloc[:10]).ultimate_.sum().sum()
    assert abs(a - b) < 1e-5
    a = bbk.ultimate_.iloc[:10].sum().sum()
    b = bbk.predict(clrd.iloc[:10], sample_weight=w.iloc[:10]).ultimate_.sum().sum()
    assert abs(a - b) < 1e-5
    a = bcc.ultimate_.iloc[:10].sum().sum()
    b = bcc.predict(clrd.iloc[:10], sample_weight=w.iloc[:10]).ultimate_.sum().sum()
    assert abs(a - b) < 1e-5

    a = bcl.ultimate_.iloc[150:153].sum().sum()
    b = bcl.predict(clrd.iloc[150:153]).ultimate_.sum().sum()
    assert abs(a - b) < 1e-5
    a = bbk.ultimate_.iloc[150:153].sum().sum()
    b = bbk.predict(clrd.iloc[150:153], sample_weight=w.iloc[150:153]).ultimate_.sum().sum()
    assert abs(a - b) < 1e-5
    a = bcc.ultimate_.iloc[150:153].sum().sum()
    b = bcc.predict(clrd.iloc[150:153], sample_weight=w.iloc[150:153]).ultimate_.sum().sum()
    assert abs(a - b) < 1e-5

    a = bcl.ultimate_.iloc[150:152].sum().sum()
    b = bcl.predict(clrd.iloc[150:152]).ultimate_.sum().sum()
    assert abs(a - b) < 1e-5
    a = bbk.ultimate_.iloc[150:152].sum().sum()
    b = bbk.predict(clrd.iloc[150:152], sample_weight=w.iloc[150:152]).ultimate_.sum().sum()
    assert abs(a - b) < 1e-5
    a = bcc.ultimate_.iloc[150:152].sum().sum()
    b = bcc.predict(clrd.iloc[150:152], sample_weight=w.iloc[150:152]).ultimate_.sum().sum()
    assert abs(a - b) < 1e-5

    a = bcl.ultimate_.iloc[150].sum().sum()
    b = bcl.predict(clrd.iloc[150]).ultimate_.sum().sum()
    assert abs(a - b) < 1e-5
    a = bbk.ultimate_.iloc[150].sum().sum()
    b = bbk.predict(clrd.iloc[150], sample_weight=w.iloc[150]).ultimate_.sum().sum()
    assert abs(a - b) < 1e-5
    a = bcc.ultimate_.iloc[150].sum().sum()
    b = bcc.predict(clrd.iloc[150], sample_weight=w.iloc[150]).ultimate_.sum().sum()
    assert abs(a - b) < 1e-5

def align_cdfs():
    ld = cl.load_sample('raa').latest_diagonal*0+40000
    model = cl.BornhuetterFerguson().fit(cl.load_sample('raa'), sample_weight=ld)
    a = model.ultimate_.iloc[..., :4, :] 
    b = model.predict(
        cl.load_sample('raa').dev_to_val().iloc[..., :4, -1].val_to_dev(),
        sample_weight=ld.iloc[..., :4, :]).ultimate_
    assert a == b
    model = cl.Chainladder().fit(cl.load_sample('raa'), sample_weight=ld)
    a = model.ultimate_.iloc[..., :4, :] 
    b = model.predict(
        cl.load_sample('raa').dev_to_val().iloc[..., :4, -1].val_to_dev(),
        sample_weight=ld.iloc[..., :4, :]).ultimate_
    assert a == b


# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from chainladder.utils.weighted_regression import WeightedRegression
from chainladder.development import Development, DevelopmentBase
import numpy as np
import pandas as pd
import warnings


class MunichAdjustment(DevelopmentBase):
    """Applies the Munich Chainladder adjustment to a set of paid/incurred
       ldfs.  The Munich method heavily relies on the ratio of paid/incurred
       and its inverse.

    Parameters
    ----------
    paid_to_incurred: tuple or list of tuples
        A tuple representing the paid and incurred ``columns`` of the triangles
        such as ``('paid', 'incurred')``
    fillna: boolean
        The MunichAdjustment will fail when P/I or I/P ratios cannot be calculated.
        Setting fillna to True will fill the triangle with expected amounts using
        the simple chainladder.

    Attributes
    ----------
    basic_cdf_: Triangle
        The univariate cumulative development patterns
    basic_sigma_: Triangle
        Sigma of the univariate ldf regression
    resids_: Triangle
        Residuals of the univariate ldf regression
    q_: Triangle
        chainladder age-to-age factors of the paid/incurred triangle and its
        inverse.  For paid measures it is (P/I) and for incurred measures it is
        (I/P).
    q_resids_: Triangle
        Residuals of q regression.
    rho_: Triangle
        Estimated conditional deviation around ``q_``
    lambda_: Series or DataFrame
        Dependency coefficient between univariate chainladder link ratios and
        ``q_resids_``
    ldf_: Triangle
        The estimated bivariate loss development patterns
    cdf_: Triangle
        The estimated bivariate cumulative development patterns

    """

    def __init__(self, paid_to_incurred=None, fillna=False):
        if type(paid_to_incurred) is dict:
            warnings.warn(
                "paid_to_incurred dict argument is deprecated, use tuple instead"
            )
            paid_to_incurred = [(k, v) for k, v in paid_to_incurred.items()]
        self.paid_to_incurred = paid_to_incurred
        self.fillna = fillna

    def fit(self, X, y=None, sample_weight=None):
        """Fit the model with X.

        Parameters
        ----------
        X : Triangle-like
            Set of LDFs to which the munich adjustment will be applied.
        y : Ignored
        sample_weight : Ignored

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        from chainladder import options

        if self.paid_to_incurred is None:
            raise ValueError("Must enter valid value for paid_to_incurred.")
        if X.array_backend == "sparse":
            obj = X.set_backend("numpy")
        else:
            obj = X.copy()
        xp = obj.get_array_module()
        self.xp = xp
        missing = xp.nan_to_num(obj.values) * obj.nan_triangle == 0
        self.rho_ = obj[obj.origin == obj.origin.min()]
        if len(xp.where(missing)[0]) > 0:
            if self.fillna:
                from chainladder.methods import Chainladder

                filler = Chainladder().fit(obj).full_expectation_
                filler = filler[filler.valuation <= obj.valuation_date].values
                obj.values = xp.where(missing, filler, obj.values)
            else:
                raise ValueError(
                    "MunichAdjustment cannot be performed when P/I or I/P "
                    + "ratios cannot be computed. Use `fillna=True` to impute zero"
                    + " values of the triangle with simple chainladder expectation."
                )

        if "ldf_" not in obj:
            obj = Development().fit_transform(obj)
        self.p_to_i_X_ = self._get_p_to_i_object(obj)
        self.p_to_i_ldf_ = self._get_p_to_i_object(obj.ldf_)
        self.p_to_i_sigma_ = self._get_p_to_i_object(obj.sigma_)
        self.q_f_, self.rho_sigma_ = self._get_MCL_model(obj)
        self.residual_, self.q_resid_ = self._get_MCL_resids(obj)
        self.lambda_coef_ = self._get_MCL_lambda(obj)
        self.ldf_ = self._set_ldf(
            obj, self._get_mcl_cdf(obj, self.munich_full_triangle_)
        )
        self.ldf_.is_cumulative = False
        self.ldf_.valuation_date = pd.to_datetime(options.ULT_VAL)
        self._map = {
            (list(X.columns).index(x)): (num % 2, num // 2)
            for num, x in enumerate(np.array(self.paid_to_incurred).flatten())
        }
        self.rho_.values = self._reshape("rho_sigma_")
        return self

    def transform(self, X):
        """ If X and self are of different shapes, align self to X, else
        return self.

        Parameters
        ----------
        X : Triangle
            The triangle to be transformed

        Returns
        -------
            X_new : New triangle with transformed attributes.
        """
        backend = X.array_backend
        if backend == "sparse":
            X_new = X.set_backend("numpy")
        else:
            X_new = X.copy()
        xp = X_new.get_array_module()
        self.xp = xp
        if "ldf_" not in X_new:
            X_new = Development().fit_transform(X_new)
        self.xp = X_new.get_array_module()
        X_new.p_to_i_X_ = self._get_p_to_i_object(X_new)
        X_new.p_to_i_ldf_ = self._get_p_to_i_object(X_new.ldf_)
        X_new.p_to_i_sigma_ = self._get_p_to_i_object(X_new.sigma_)
        X_new.q_f_, X_new.rho_sigma_ = self._get_MCL_model(X_new)
        X_new.munich_full_triangle_ = self._get_munich_full_triangle_(
            X_new.p_to_i_X_,
            X_new.p_to_i_ldf_,
            X_new.p_to_i_sigma_,
            self.lambda_coef_,
            X_new.rho_sigma_,
            X_new.q_f_,
        )
        X_new.ldf_ = self._set_ldf(
            X_new, self._get_mcl_cdf(X_new, X_new.munich_full_triangle_)
        )
        del self.xp
        triangles = ["rho_", "lambda_", "lambda_coef_"]
        for item in triangles:
            setattr(X_new, item, getattr(self, item))
        X_new._set_slicers()
        X_new.sigma_ = X_new.std_err_ = X_new.ldf_ * 0 + 1
        return X_new

    def _get_p_to_i_object(self, obj):
        if type(self.paid_to_incurred) is tuple:
            p_to_i = [self.paid_to_incurred]
        else:
            p_to_i = self.paid_to_incurred
        xp = obj.get_array_module()
        paid = obj[[item[0] for item in p_to_i][0]]
        for item in [item[0] for item in p_to_i][1:]:
            paid[item] = obj[item]
        incurred = obj[[item[1] for item in p_to_i][0]]
        for item in [item[1] for item in p_to_i][1:]:
            incurred[item] = obj[item]
        paid = paid.values[None]
        incurred = incurred.values[None]
        return xp.concatenate((paid, incurred), axis=0)

    def _p_to_i_concate(self, obj_p, obj_i, xp):
        return xp.concatenate((obj_p[None], obj_i[None]), 0)

    def _get_MCL_model(self, X):
        xp = X.get_array_module()
        p, i = self.p_to_i_X_[0], self.p_to_i_X_[1]
        modelsP = WeightedRegression(axis=2, thru_orig=True, xp=xp)
        modelsP = modelsP.fit(p, i, 1 / p).sigma_fill(X.sigma_interpolation)
        modelsI = WeightedRegression(axis=2, thru_orig=True, xp=xp)
        modelsI = modelsI.fit(i, p, 1 / i).sigma_fill(X.sigma_interpolation)
        q_f = self._p_to_i_concate(modelsP.slope_, modelsI.slope_, xp)
        rho_sigma = self._p_to_i_concate(modelsP.sigma_, modelsI.sigma_, xp)
        return xp.swapaxes(q_f, -1, -2), xp.swapaxes(rho_sigma, -1, -2)

    def _get_MCL_resids(self, X):
        xp = X.get_array_module()
        p_to_i_ata = self._get_p_to_i_object(X.link_ratio)
        p_to_i_ldf = self.p_to_i_ldf_
        p_to_i_sigma = self.p_to_i_sigma_
        paid, incurred = self.p_to_i_X_[0], self.p_to_i_X_[1]
        p_to_i_ldf = xp.unique(p_to_i_ldf, axis=-2)  # May cause issues later
        p_to_i_sigma = xp.unique(p_to_i_sigma, axis=-2)  # May cause issues
        residP = (
            (p_to_i_ata[0] - p_to_i_ldf[0])
            / p_to_i_sigma[0]
            * xp.sqrt(paid[..., :-1, :-1])
        )
        residI = (
            (p_to_i_ata[1] - p_to_i_ldf[1])
            / p_to_i_sigma[1]
            * xp.sqrt(incurred[..., :-1, :-1])
        )
        nans = (X - X[X.valuation == X.valuation_date]).values[0, 0] * 0 + 1
        q_resid = (
            (paid / incurred - self.q_f_[1])
            / self.rho_sigma_[1]
            * xp.sqrt(incurred)
            * nans
        )
        q_inv_resid = (
            (incurred / paid - 1 / self.q_f_[1])
            / self.rho_sigma_[0]
            * xp.sqrt(paid)
            * nans
        )
        resid = self._p_to_i_concate(residP, residI, xp)
        q_resid = self._p_to_i_concate(q_inv_resid, q_resid, xp)
        return resid, q_resid

    def _get_MCL_lambda(self, obj):
        xp = obj.get_array_module()
        k, v, o, d = self.residual_[1].shape
        w = xp.reshape(self.residual_[1], (k, v, o * d))
        w[w == 0] = xp.nan
        w = w * 0 + 1
        lambdaI = (
            WeightedRegression(thru_orig=True, axis=-1, xp=xp)
            .fit(
                xp.reshape(self.q_resid_[1][..., :-1, :-1], (k, v, o * d)),
                xp.reshape(self.residual_[1], (k, v, o * d)),
                w,
            )
            .slope_
        )
        lambdaP = (
            WeightedRegression(thru_orig=True, axis=-1, xp=xp)
            .fit(
                xp.reshape(self.q_resid_[0][..., :-1, :-1], (k, v, o * d)),
                xp.reshape(self.residual_[0], (k, v, o * d)),
                w,
            )
            .slope_
        )
        return self._p_to_i_concate(lambdaP, lambdaI, xp)[..., None]

    @property
    def munich_full_triangle_(self):
        return self._get_munich_full_triangle_(
            self.p_to_i_X_,
            self.p_to_i_ldf_,
            self.p_to_i_sigma_,
            self.lambda_coef_,
            self.rho_sigma_,
            self.q_f_,
        )

    def _get_munich_full_triangle_(
        self, p_to_i_X_, p_to_i_ldf_, p_to_i_sigma_, lambda_coef_, rho_sigma_, q_f_
    ):
        xp = self.xp if hasattr(self, "xp") else self.ldf_.get_array_module()
        full_paid = xp.nan_to_num(p_to_i_X_[0][..., 0:1])
        full_incurred = p_to_i_X_[1][..., 0:1]
        for i in range(p_to_i_X_[0].shape[-1] - 1):
            paid = (
                p_to_i_ldf_[0][..., i : i + 1]
                + lambda_coef_[0]
                * p_to_i_sigma_[0][..., i : i + 1]
                / rho_sigma_[0][..., i : i + 1]
                * (
                    full_incurred[..., -1:] / full_paid[..., -1:]
                    - q_f_[0][..., i : i + 1]
                )
            ) * full_paid[..., -1:]
            inc = (
                p_to_i_ldf_[1][..., i : i + 1]
                + self.lambda_coef_[1]
                * p_to_i_sigma_[1][..., i : i + 1]
                / rho_sigma_[1][..., i : i + 1]
                * (
                    full_paid[..., -1:] / full_incurred[..., -1:]
                    - q_f_[1][..., i : i + 1]
                )
            ) * full_incurred[..., -1:]
            full_incurred = xp.concatenate(
                (
                    full_incurred,
                    xp.nan_to_num(p_to_i_X_[1][..., i + 1 : i + 2])
                    + (1 - xp.nan_to_num(p_to_i_X_[1][..., i + 1 : i + 2] * 0 + 1))
                    * inc,
                ),
                axis=3,
            )
            full_paid = xp.concatenate(
                (
                    full_paid,
                    xp.nan_to_num(p_to_i_X_[0][..., i + 1 : i + 2])
                    + (1 - xp.nan_to_num(p_to_i_X_[0][..., i + 1 : i + 2] * 0 + 1))
                    * paid,
                ),
                axis=3,
            )
        return self._p_to_i_concate(full_paid, full_incurred, xp)

    def _get_mcl_cdf(self, X, munich_full_triangle_):
        """ needs to be an attribute that gets assigned.  requires we overwrite
            the cdf and ldf methods with
        """
        xp = X.get_array_module()
        obj = X.cdf_.copy()
        obj.values = xp.repeat(obj.values, len(X.odims), 2)
        obj.odims = X.odims
        if type(self.paid_to_incurred) is tuple:
            p_to_i = [self.paid_to_incurred]
        else:
            p_to_i = self.paid_to_incurred
        cdf_triangle = munich_full_triangle_
        cdf_triangle = cdf_triangle[..., -1:] / cdf_triangle[..., :-1]
        paid = [item[0] for item in p_to_i]
        for n, item in enumerate(paid):
            idx = np.where(X.cdf_.vdims == item)[0][0]
            obj.values[:, idx : idx + 1, ...] = cdf_triangle[0, :, n : n + 1, ...]
        incurred = [item[1] for item in p_to_i]
        for n, item in enumerate(incurred):
            idx = np.where(X.cdf_.vdims == item)[0][0]
            obj.values[:, idx : idx + 1, ...] = cdf_triangle[1, :, n : n + 1, ...]
        obj._set_slicers()
        return obj

    def _set_ldf(self, X, cdf):
        ldf_tri = cdf.values.copy()
        xp = X.get_array_module()
        ldf_tri = xp.concatenate((ldf_tri, xp.ones(ldf_tri.shape)[..., -1:]), -1)
        ldf_tri = ldf_tri[..., :-1] / ldf_tri[..., 1:]
        obj = cdf.copy()
        obj.values = ldf_tri
        obj.ddims = X.link_ratio.ddims
        obj.is_pattern = True
        obj.is_cumulative = False
        obj._set_slicers
        return obj

    def _reshape(self, measure):
        xp = self.xp if hasattr(self, "xp") else self.ldf_.get_array_module()
        map = self._map
        return xp.concatenate(
            [
                getattr(self, measure)[map[k][0], :, map[k][1] : map[k][1] + 1, ...]
                for k in range(len(map))
            ],
            axis=1,
        )

    @property
    def lambda_(self):
        obj = self.ldf_.copy()
        obj.odims = obj.odims[0:1]
        obj.ddims = obj.ddims[0:1]
        obj.values = self._reshape("lambda_coef_")
        return obj.to_frame(origin_as_datetime=False)

    @property
    def basic_cdf_(self):
        obj = self.ldf_.copy()
        obj.values = self._reshape("p_to_i_ldf_")
        return obj

    @property
    def basic_sigma_(self):
        obj = self.ldf_.copy()
        obj.values = self._reshape("p_to_i_sigma_")
        return obj

    @property
    def resids_(self):
        obj = self.ldf_.copy()
        obj.values = self._reshape("residual_")
        obj.odims = self.cdf_.odims[: obj.values.shape[2]]
        return obj

    @property
    def q_(self):
        obj = self.rho_.copy()
        obj.odims = self.cdf_.odims
        obj.values = self._reshape("q_f_")
        return obj

    @property
    def q_resids_(self):
        obj = self.ldf_.copy()
        obj.values = self._reshape("q_resid_")[
            ..., : self.residual_.shape[-2], : self.residual_.shape[-1]
        ]
        obj.odims = obj.odims[: obj.values.shape[2]]
        obj.ddims = obj.ddims[: obj.values.shape[3]]
        return obj



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from chainladder.methods.chainladder import Chainladder
from sklearn.base import BaseEstimator, TransformerMixin
import numpy as np
import copy
import warnings
from chainladder.core.io import EstimatorIO


class BerquistSherman(BaseEstimator, TransformerMixin, EstimatorIO):
    """
    Class to alter the inner diagonals of a Triangle using the methods described
    by Berquist and Sherman.

    Parameters
    ----------
    paid_amount: str
        The triangle column associated with paid amounts
    incurred_amount: str
        The triangle column associated with incurred amounts
    reported_count: str
        The triangle column associated with reported claim count
    closed_count: str
        The triangle column associated with closed claim counts
    trend: float (default=0.0)
        Trend rate underlying average open case reserves.
    reported_count_estimator: Estimator
        An IBNR estimator for reported_count used to calculate closed_count
        disposal rates.  Estimator can be a Pipeline.  If None selected then
        basic Chainladder model is used.

    Attributes
    ----------
    adjusted_triangle_: Triangle
        A set of triangles represented by each simulation
    disposal_rate_: Triangle
        The disposal rates of closed claims based on the reported_count_estimator
    a_: Triangle
        Two-period Exponential intercept parameters
    b_: Triangle
        Two-period Exponential slope parameters
    """

    def __init__(
        self,
        paid_amount=None,
        incurred_amount=None,
        reported_count=None,
        closed_count=None,
        trend=0.0,
        reported_count_estimator=None,
    ):
        self.paid_amount = paid_amount
        self.incurred_amount = incurred_amount
        self.reported_count = reported_count
        self.closed_count = closed_count
        self.trend = trend
        self.reported_count_estimator = reported_count_estimator

    def fit(self, X, y=None, sample_weight=None):
        backend = X.array_backend
        if backend == "sparse":
            obj = X.set_backend("numpy")
        else:
            obj = X.copy()
        xp = obj.get_array_module()
        if not (
            self.paid_amount in X.columns
            and self.incurred_amount in X.columns
            and self.reported_count in X.columns
            and self.closed_count in X.columns
        ):
            raise ValueError(
                "Must enter values valid columns for paid_amount, incurred_amount, reported_count and closed_count"
            )
        paid_amount = self.paid_amount
        incurred_amount = self.incurred_amount
        reported_count = self.reported_count
        closed_count = self.closed_count
        reported_count_estimator = self.reported_count_estimator
        # Case reserve adequacy adjustment
        open_count = obj[reported_count] - obj[closed_count]
        avg_case = (obj[incurred_amount] - obj[paid_amount]) / open_count
        adj_avg_case = (
            avg_case.trend(1 / (1 + self.trend) - 1, axis="valuation")
            / avg_case
            * avg_case[avg_case.valuation == avg_case.valuation_date].sum("origin")
        )
        adj_incurred_amount = adj_avg_case * open_count + obj[paid_amount]

        # Paid and closed claim adjustments
        if reported_count_estimator is None:
            reported_count_estimator = Chainladder()
        reported_count_estimator.fit(obj[reported_count])
        if reported_count_estimator.__class__.__name__ == "Pipeline":
            rep_cnt_ult = reported_count_estimator.named_steps[
                reported_count_estimator.steps[-1][0]
            ].ultimate_
        else:
            rep_cnt_ult = reported_count_estimator.ultimate_
        disposal_rate = obj[closed_count] / rep_cnt_ult
        adj_closed_clm = (
            (disposal_rate * 0 + 1)
            * disposal_rate[disposal_rate.valuation == X.valuation_date].mean("origin")
            * rep_cnt_ult
        )
        adj_closed_clm.valuation_date = disposal_rate.valuation_date = X.valuation_date

        # Two-period exponential regression
        y = obj[paid_amount]
        x = obj[closed_count]
        x0 = x.values[..., :-1]
        x1 = x.values[..., 1:]
        y0 = xp.log(y.values[..., :-1])
        y1 = xp.log(y.values[..., 1:])
        b = ((x0 * y0 + x1 * y1) / 2 - (x0 + x1) / 2 * (y0 + y1) / 2) / (
            (x0 ** 2 + x1 ** 2) / 2 - ((x0 + x1) / 2) ** 2
        )
        a = np.exp((y0 + y1) / 2 - b * (x0 + x1) / 2)

        # Need to Ignore warnings for NaN cells
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            lookup = (
                np.maximum(
                    xp.concatenate(
                        [
                            (
                                adj_closed_clm.values[..., i : i + 1]
                                > obj[closed_count].values
                            ).sum(axis=-1, keepdims=True)
                            for i in range(adj_closed_clm.shape[-1])
                        ],
                        axis=-1,
                    ),
                    1,
                )
                - 1
            )
        a = (
            xp.concatenate(
                [
                    a[..., i, lookup[0, 0, i : i + 1, :]]
                    for i in range(lookup.shape[-2])
                ],
                -2,
            )
            * adj_closed_clm.nan_triangle[None, None, ...]
        )
        b = (
            xp.concatenate(
                [
                    b[..., i, lookup[0, 0, i : i + 1, :]]
                    for i in range(lookup.shape[-2])
                ],
                -2,
            )
            * adj_closed_clm.nan_triangle[None, None, ...]
        )
        # Adjust paids
        adj_paid_claims = adj_closed_clm * 0 + xp.exp(adj_closed_clm.values * b) * a
        adj_paid_claims = (
            y[y.valuation == y.valuation_date]
            + adj_paid_claims[
                adj_paid_claims.valuation < adj_paid_claims.valuation_date
            ]
        )

        adjusted_triangle_ = copy.deepcopy(obj)
        adjusted_triangle_[paid_amount] = adj_paid_claims
        adjusted_triangle_[incurred_amount] = adj_incurred_amount
        adjusted_triangle_[closed_count] = adj_closed_clm
        adjusted_triangle_ = adjusted_triangle_[
            adjusted_triangle_.valuation <= obj.valuation_date
        ]
        self.adjusted_triangle_ = adjusted_triangle_.set_backend(backend)
        self.disposal_rate_ = disposal_rate.set_backend(backend)
        self.a_ = a
        self.b_ = b
        return self

    def transform(self, X):
        """ If X and self are of different shapes, align self to X, else
        return self.

        Parameters
        ----------
        X: Triangle
            The triangle to be transformed

        Returns
        -------
            X_new: New triangle with transformed attributes.
        """
        X_new = copy.deepcopy(X)
        X_new[self.paid_amount] = self.adjusted_triangle_[self.paid_amount]
        X_new[self.incurred_amount] = self.adjusted_triangle_[self.incurred_amount]
        X_new[self.reported_count] = self.adjusted_triangle_[self.reported_count]
        X_new[self.closed_count] = self.adjusted_triangle_[self.closed_count]
        X_new.a_ = self.a_
        X_new.b_ = self.b_
        return X_new

    def set_params(self, **params):
        from chainladder.utils.utility_functions import read_json

        if type(params["reported_count_estimator"]) is str:
            params["reported_count_estimator"] = read_json(
                params["reported_count_estimator"]
            )
        return super().set_params(**params)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import pandas as pd
import numpy as np
from chainladder.utils.cupy import cp
from chainladder.utils.sparse import sp
from chainladder.utils.dask import dp
import warnings

from chainladder.core.display import TriangleDisplay
from chainladder.core.dunders import TriangleDunders
from chainladder.core.pandas import TrianglePandas
from chainladder.core.slice import TriangleSlicer
from chainladder.core.io import TriangleIO
from chainladder.core.common import Common
from chainladder import options
from chainladder.utils.utility_functions import num_to_nan, concat


class TriangleBase(TriangleIO, TriangleDisplay, TriangleSlicer,
                   TriangleDunders, TrianglePandas, Common):
    """ This class handles the initialization of a triangle """

    @property
    def shape(self):
        return self.values.shape

    @staticmethod
    def _input_validation(data, index, columns, origin, development):
        """ Validate/sanitize inputs """
        def str_to_list(arg):
            if arg is None:
                return
            if type(arg) in [str, pd.Period]:
                return [arg]
            else:
                return list(arg)
        index = str_to_list(index)
        columns = str_to_list(columns)
        origin = str_to_list(origin)
        development = str_to_list(development)
        if "object" in data[columns].dtypes:
            raise TypeError("column attribute must be numeric.")
        if data[columns].shape[1] != len(columns):
            raise AttributeError("Columns are required to have unique names")
        return index, columns, origin, development

    @staticmethod
    def _set_development(data, development, development_format, origin_date):
        """ Initialize development and its grain """
        if development:
            development_date = TriangleBase._to_datetime(
                data, development, period_end=True, format=development_format)
            #if np.all(development_date.dt.strftime('%m-%d') == '01-01'):
            #    development_date = pd.Series(pd.PeriodIndex(development_date, freq='A').to_timestamp(how='e'))
        else:
            o_max = pd.Period(
                origin_date.max(),
                freq=TriangleBase._get_grain(origin_date)).to_timestamp(how='e')
            development_date = pd.Series([o_max] * len(origin_date))
        development_date.name = "__development__"
        if pd.Series(development_date).dt.year.min() == \
           pd.Series(development_date).dt.year.max() == 1970:
            raise ValueError(
                'Development lags could not be determined. This may be because development'
                'is expressed as an age where a date-like vector is required')
        return development_date

    @staticmethod
    def _set_index(col, unique):
        return col.map(dict(zip(unique, range(len(unique))))).values[None].T

    @staticmethod
    def _aggregate_data(data, origin_date, development_date, index, columns):
        """ Summarize dataframe to the level specified in axes """
        if type(data) != pd.DataFrame:
            # Dask dataframes are mutated
            data["__origin__"] = origin_date
            data["__development__"] = development_date
            key_gr = ["__origin__", "__development__"] + [
                data[item] for item in ([] if not index else index)]
            data_agg = data.groupby(key_gr)[columns].sum().reset_index().fillna(0)
            data = data.drop(['__origin__', '__development__'], axis=1)
        else:
            # Summarize dataframe to the level specified in axes
            key_gr = [origin_date, development_date] + [
                data[item] for item in ([] if not index else index)]
            data_agg = data[columns].groupby(key_gr).sum().reset_index().fillna(0)
            data_agg["__origin__"] = data_agg[origin_date.name]
            data_agg["__development__"] = data_agg[development_date.name]
        # origin <= development is required - truncate bad records if not true
        valid = data_agg["__origin__"] <= data_agg["__development__"]
        if sum(~valid) > 0:
            warnings.warn("""
                Observations with development before
                origin start have been removed.""")
            valid = valid.compute() if hasattr(valid, 'compute') else valid
            data_agg = data_agg[valid]
        return data_agg

    @staticmethod
    def _set_kdims(data_agg, index):
        kdims = data_agg[index].drop_duplicates().reset_index(drop=True).reset_index()
        key_idx = (data_agg[index].merge(
            kdims, how="left", on=index)["index"].values[None].T)
        return kdims.drop("index", axis=1).values, key_idx

    @staticmethod
    def _set_odims(data_agg, date_axes):
        odims = np.sort(date_axes["__origin__"].unique())
        orig_idx = TriangleBase._set_index(data_agg["__origin__"], odims)
        return odims, orig_idx

    @staticmethod
    def _set_ddims(data_agg, date_axes):
        if date_axes["__development__"].nunique() > 1:
            dev_lag = TriangleBase._development_lag(
                data_agg["__origin__"], data_agg["__development__"])
            ddims = np.sort(
                TriangleBase._development_lag(
                    date_axes["__origin__"], date_axes["__development__"]
                ).unique())
            dev_idx = TriangleBase._set_index(dev_lag, ddims)
        else:
            ddims = pd.DatetimeIndex(
                [data_agg['__development__'].max()], name='valuation')
            dev_idx = np.zeros((len(data_agg), 1))
        return ddims, dev_idx

    @staticmethod
    def _set_values(data_agg, key_idx, columns, orig_idx, dev_idx):
        val_idx = (
            ((np.ones(len(data_agg))[None].T) * range(len(columns)))
            .reshape((1, -1), order="F").T)
        coords = np.concatenate(
            tuple([np.concatenate((orig_idx, dev_idx), 1)] * len(columns)), 0)
        coords = np.concatenate(
            (np.concatenate(tuple([key_idx] * len(columns)), 0), val_idx, coords), 1)
        amts = np.concatenate(
            [data_agg[col].fillna(0).values for col in data_agg[columns]]).astype("float64")
        return coords.T.astype('int64'), amts

    def _len_check(self, x, y):
        if len(x) != len(y):
            raise ValueError(
                "Length mismatch: Expected axis has",
                "{} elements, new values have".format(len(x)),
                "{} elements".format(len(y)),
            )

    def _get_date_axes(self, origin_date, development_date, origin_grain, development_grain):
        """ Function to find any missing origin dates or development dates that
            would otherwise mess up the origin/development dimensions.
        """
        o = pd.period_range(
            start=origin_date.min(), end=origin_date.max(),
            freq=origin_grain).to_timestamp(how='s')
        d = pd.period_range(
            start=development_date.min(), end=development_date.max(),
            freq=development_grain).to_timestamp(how='e')
        c = pd.DataFrame(TriangleBase._cartesian_product(o, d),
                         columns=['__origin__', '__development__'])
        return c[c['__development__']>c['__origin__']]


    @property
    def nan_triangle(self):
        """Given the current triangle shape and valuation, it determines the
           appropriate placement of NANs in the triangle for future valuations.
           This becomes useful when managing array arithmetic.
        """
        xp = self.get_array_module()
        if min(self.values.shape[2:]) == 1:
            return xp.ones(self.values.shape[2:], dtype="float16")
        val_array = np.array(self.valuation).reshape(self.shape[-2:], order="f")
        nan_triangle = np.array(pd.DataFrame(val_array) > self.valuation_date)
        nan_triangle = xp.array(np.where(nan_triangle, xp.nan, 1), dtype="float16")
        return nan_triangle

    @staticmethod
    def _to_datetime(data, fields, period_end=False, format=None):
        """For tabular form, this will take a set of data
        column(s) and return a single date array.  This function heavily
        relies on pandas, but does two additional things:
        1. It extends the automatic inference using date_inference_list
        2. it allows pd_to_datetime on a set of columns
        """
        # Concat everything into one field
        if len(fields) > 1:
            target_field = data[fields].astype(str).apply(lambda x: "-".join(x), axis=1)
        else:
            target_field = data[fields].iloc[:, 0]
        if hasattr(target_field, "dt"):
            target = target_field
            if type(target.iloc[0]) == pd.Period:
                return target.dt.to_timestamp(how={1: "e", 0: "s"}[period_end])
        else:
            datetime_arg = target_field.unique()
            format = [{"arg": datetime_arg, "format": format}] if format else []
            date_inference_list = format + [
                {"arg": datetime_arg, "format": "%Y%m"},
                {"arg": datetime_arg, "format": "%Y"},
                {"arg": datetime_arg, "infer_datetime_format": True},
            ]
            for item in date_inference_list:
                try:
                    arr = dict(zip(datetime_arg, pd.to_datetime(**item)))
                    break
                except:
                    pass
            target = target_field.map(arr)
        return target

    @staticmethod
    def _development_lag(origin, development):
        """ For tabular format, this will convert the origin/development
            difference to a development lag """
        return ((development - origin) /
                np.timedelta64(1, 'M')).round(0).astype(int)

    @staticmethod
    def _get_grain(dates, trailing=False, kind='origin'):
        """ Determines Grain of origin or valuation vector
        
        Parameters:
        
        dates: pd.Series[datetime64[ns]]
            A Datetime Series
        trailing:
            Set to False if you want to treat December as period end. Set
            to True if you want it inferred from the data.
        """
        months = dates.dt.month.unique()
        diffs = np.diff(np.sort(months))
        if len(dates.unique()) == 1:
            grain = 'M'
        elif len(months) == 1:
            grain = 'A'
        elif np.all(diffs == 6):
            grain = '2Q'
        elif np.all(diffs == 3):
            grain = 'Q'
        else:
            grain = 'M'
        if trailing and grain != 'M':
            if kind == 'origin':
                end = (dates.min() - pd.DateOffset(days=1)).strftime('%b').upper()
            else:
                # If inferred to beginning of calendar period, 1/1 from YYYY, 4/1 from YYYYQQ
                if dates.dt.strftime('%m%d').isin(['0101', '0401', '0701', '1001']).any():
                    end = (dates.min() - pd.DateOffset(days=1, years=-1)).strftime('%b').upper()
                else:
                    end = dates.max().strftime('%b').upper()
            grain = grain + '-' + end
        return grain

    @staticmethod
    def _cartesian_product(*arrays):
        """A fast implementation of cartesian product, used for filling in gaps
        in triangles (if any)"""
        arr = np.empty(
            [len(a) for a in arrays] + [len(arrays)],
            dtype=np.result_type(*arrays))
        for i, a in enumerate(np.ix_(*arrays)):
            arr[..., i] = a
        arr = arr.reshape(-1, len(arrays))
        return arr

    def get_array_module(self, arr=None):
        backend = (self.array_backend if arr is None
                   else arr.__class__.__module__.split(".")[0])
        modules = {"cupy": cp, "sparse": sp, "numpy": np, "dask": dp}
        if modules.get(backend, None):
            return modules.get(backend, None)
        else:
            raise ValueError("Array backend is invalid or not properly set.")

    def _auto_sparse(self):
        """ Auto sparsifies at 30Mb or more and 20% density or less """
        if not options.AUTO_SPARSE:
            return self
        n = np.prod(list(self.shape) + [8 / 1e6])
        if (self.array_backend == "numpy" and n > 30
            and 1 - np.isnan(self.values).sum() / n * (8 / 1e6) < 0.2):
            self.set_backend("sparse", inplace=True)
        if self.array_backend == "sparse" and not (
           self.values.density < 0.2 and n > 30):
            self.set_backend("numpy", inplace=True)
        return self


    @property
    def valuation(self):
        ddims = self.ddims
        if self.is_val_tri:
            out = pd.DataFrame(np.repeat(self.ddims.values[None],
                               len(self.odims), 0))
            return pd.DatetimeIndex(out.unstack().values)
        ddim_arr = ddims - ddims[0]
        origin = np.minimum(self.odims, np.datetime64(self.valuation_date))
        val_array = origin.astype("datetime64[M]") + np.timedelta64(ddims[0], "M")
        val_array = val_array.astype("datetime64[ns]") - np.timedelta64(1, "ns")
        val_array = val_array[:, None]
        s = slice(None, -1) if ddims[-1] == 9999 else slice(None, None)
        val_array = (
            val_array.astype("datetime64[M]") + ddim_arr[s][None, :] + 1
        ).astype("datetime64[ns]") - np.timedelta64(1, "ns")
        if ddims[-1] == 9999:
            ult = np.repeat(np.datetime64(options.ULT_VAL), val_array.shape[0])[:, None]
            val_array = np.concatenate((val_array, ult,), axis=1,)
        return pd.DatetimeIndex(val_array.reshape(1, -1, order="F")[0])

    def _drop_subtriangles(self):
        """ Removes subtriangles from a Triangle instance """
        sub_tris = [k for k, v in vars(self).items()
                    if isinstance(v, TriangleBase)]
        if 'ldf_' in sub_tris:
            del self.ldf_
        if 'sigma_' in sub_tris:
            del self.sigma_
        if 'std_err_' in sub_tris:
            del self.std_err_

    @property
    def subtriangles(self):
        """ Lists subtriangles from a Triangle instance """
        return  [k for k, v in vars(self).items() if isinstance(v, TriangleBase)]

    def __array__(self):
        return self.values

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        obj = self.copy()
        if method == '__call__':
            inputs = [i.values if hasattr(i, 'values') else i for i in inputs]
            obj.values = ufunc(*inputs, **kwargs)
            return obj
        else:
            raise NotImplementedError()

    def __array_function__(self, func, types, args, kwargs):
        from chainladder.utils.utility_functions import concat
        methods_as_funcs = list(set(dir(np)).intersection(set(dir(self))) -
                                {'__dir__', '__doc__'})
        methods_as_funcs = {getattr(np, i): getattr(self, i)
                            for i in methods_as_funcs}
        HANDLED_FUNCTIONS = {np.concatenate: concat, np.round: self.__round__}
        HANDLED_FUNCTIONS = {**HANDLED_FUNCTIONS, **methods_as_funcs}
        if func not in HANDLED_FUNCTIONS:
            return NotImplemented
        if not all(issubclass(t, self.__class__) for t in types):
            return NotImplemented
        if func in methods_as_funcs:
            args = args[1:]
        return HANDLED_FUNCTIONS[func](*args, **kwargs)

    def compute(self, *args, **kwargs):
        if hasattr(self.values, 'chunks'):
            obj = self.copy()
            obj.values = obj.values.compute(*args, **kwargs)
            m = obj.get_array_module(obj.values)
            if m == sp:
                obj.array_backend = 'sparse'
            if m == cp:
                obj.array_backend = 'cupy'
            if m == np:
                obj.array_backend = 'numpy'
            return obj
        return self


def is_chainladder(estimator):
    """Return True if the given estimator is a chainladder based method.
    Parameters
    ----------
    estimator : object
        Estimator object to test.
    Returns
    -------
    out : bool
        True if estimator is a chainladder based method and False otherwise.
    """
    return getattr(estimator, "_estimator_type", None) == "chainladder"



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from scipy.stats import binom, norm, rankdata
from scipy.special import comb
import pandas as pd
import numpy as np


class DevelopmentCorrelation:
    """
    Mack (1997) test for correlations between subsequent development
    factors. Results should be within confidence interval range
    otherwise too much correlation

    Parameters
    ----------
    triangle: Triangle
        Triangle on which to estimate correlation between subsequent development
        factors.
    p_critical: float (default=0.10)
        Value between 0 and 1 representing the confidence level for the test. A
        value of 0.1 implies 90% confidence.

    Attributes
    ----------
    t_critical: DataFrame
        Boolean value for whether correlation is too high based on ``p_critical``
        confidence level.
    t_expectation: DataFrame
        Values representing the Spearman rank correlation
    t_variance: float
        Variance measure of Spearman rank correlation
    range: tuple
        Range within which ``t_expectation`` must fall for independence assumption
        to be significant.
    """

    def __init__(self, triangle, p_critical=0.5):
        self.p_critical = p_critical
        if triangle.array_backend != "numpy":
            triangle = triangle.set_backend("numpy")
        xp = triangle.get_array_module()

        m1 = triangle.link_ratio
        m1_val = xp.apply_along_axis(rankdata, 2, m1.values) * (m1.values * 0 + 1)
        m2 = triangle[triangle.valuation < triangle.valuation_date].link_ratio
        m2.values = xp.apply_along_axis(rankdata, 2, m2.values) * (m2.values * 0 + 1)
        m1 = m2.copy()
        m1.values = m1_val[..., : m2.shape[2], 1:]
        numerator = ((m1 - m2) ** 2).sum("origin")
        numerator.values = numerator.values[..., :-1]
        numerator.ddims = numerator.ddims[:-1]
        I = triangle.shape[3]
        k = xp.array(range(2, 2 + numerator.shape[3]))
        denominator = ((I - k) ** 3 - I + k)[None, None, None]
        self.t = 1 - 6 * xp.nan_to_num(numerator.values) / denominator
        weight = (I - k - 1)[None, None, None]
        t_expectation = (
            xp.sum(xp.nan_to_num(weight * self.t), axis=3) / xp.sum(weight, axis=3)
        )[..., None]
        idx = triangle.index.set_index(triangle.key_labels).index
        self.t_variance = 2 / ((I - 2) * (I - 3))
        self.t = pd.DataFrame(self.t[..., 0, 0], columns=triangle.vdims, index=idx)
        self.t_expectation = pd.DataFrame(
            t_expectation[..., 0, 0], columns=triangle.vdims, index=idx
        )
        self.range = (
            norm.ppf(0.5 - (1 - p_critical) / 2) * xp.sqrt(self.t_variance),
            norm.ppf(0.5 + (1 - p_critical) / 2) * xp.sqrt(self.t_variance),
        )
        self.t_critical = (self.t_expectation < self.range[0]) | (
            self.t_expectation > self.range[1]
        )


class ValuationCorrelation:
    """
    Mack (1997) test for calendar year effect.A calendar period has impact
    across developments if the probability of the number of small (or large)
    development factors, Z, in that period occurring randomly is less than
    ``p_critical``

    Parameters
    ----------
    triangle: Triangle
        Triangle on which to test whether the calendar effects violate independence
        requirements of the chainladder method.
    p_critical: float (default=0.10)
        Value between 0 and 1 representing the confidence level for the test. 0.1
        implies 90% confidence.
    total: boolean
        Whether to calculate valuation correlation in total across all
        years (True) consistent with Mack 1993 or for each year separately
        (False) consistent with Mack 1997.

    Attributes
    ----------
    z : Triangle or DataFrame
        Z values for each Valuation Period
    z_critical : Triangle or DataFrame
        Boolean value for whether correlation is too high based on ``p_critical``
        confidence level.
    z_expectation : Triangle or DataFrame
        The expected value of Z.
    z_variance : Triangle or DataFrame
        The variance value of Z.
    """

    def __init__(self, triangle, p_critical=0.1, total=True):
        def pZlower(z, n, p=0.5):
            return min(1, 2 * binom.cdf(z, n, p))

        self.p_critical = p_critical
        self.total = total
        triangle = triangle.set_backend("numpy")
        xp = triangle.get_array_module()
        lr = triangle.link_ratio
        m1 = xp.apply_along_axis(rankdata, 2, lr.values) * (lr.values * 0 + 1)
        med = xp.nanmedian(m1, axis=2, keepdims=True)
        m1large = (xp.nan_to_num(m1) > med) + (lr.values * 0)
        m1small = (xp.nan_to_num(m1) < med) + (lr.values * 0)
        m2large = triangle.link_ratio
        m2large.values = m1large
        m2small = triangle.link_ratio
        m2small.values = m1small
        S = xp.nan_to_num(m2small.dev_to_val().sum(axis=2).set_backend('numpy').values)
        L = xp.nan_to_num(m2large.dev_to_val().sum(axis=2).set_backend('numpy').values)
        z = xp.minimum(L, S)
        n = L + S
        m = xp.floor((n - 1) / 2)
        c = comb(n - 1, m)
        EZ = (n / 2) - c * n / (2 ** n)
        VarZ = n * (n - 1) / 4 - c * n * (n - 1) / (2 ** n) + EZ - EZ ** 2
        if not self.total:
            T = []
            for i in range(0, xp.max(m1large.shape[2:]) + 1):
                T.append(
                    [
                        pZlower(i, j, 0.5)
                        for j in range(0, xp.max(m1large.shape[2:]) + 1)
                    ]
                )
            T = np.array(T)
            z_idx, n_idx = z.astype(int), n.astype(int)
            self.probs = T[z_idx, n_idx]
            z_critical = triangle[triangle.valuation > triangle.valuation.min()]
            z_critical = z_critical.dev_to_val().dropna().sum("origin") * 0
            z_critical.values = np.array(self.probs) < p_critical
            z_critical.odims = triangle.odims[0:1]
            self.z_critical = z_critical
            self.z = self.z_critical.copy()
            self.z.values = z
            self.z_expectation = self.z_critical.copy()
            self.z_expectation.values = EZ
            self.z_variance = self.z_critical.copy()
            self.z_variance.values = VarZ
        else:
            ci2 = norm.ppf(0.5 - (1 - p_critical) / 2) * xp.sqrt(xp.sum(VarZ, axis=-1))
            self.range = (xp.sum(VarZ, axis=-1) + ci2, xp.sum(VarZ, axis=-1) - ci2)
            idx = triangle.index.set_index(triangle.key_labels).index
            self.z_critical = pd.DataFrame(
                (
                    (self.range[0] > VarZ.sum(axis=-1))
                    | (VarZ.sum(axis=-1) > self.range[1])
                )[..., 0],
                columns=triangle.vdims,
                index=idx,
            )
            self.z = pd.DataFrame(
                z.sum(axis=-1)[..., 0], columns=triangle.vdims, index=idx
            )
            self.z_expectation = pd.DataFrame(
                EZ.sum(axis=-1)[..., 0], columns=triangle.vdims, index=idx
            )
            self.z_variance = pd.DataFrame(
                VarZ.sum(axis=-1)[..., 0], columns=triangle.vdims, index=idx
            )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from setuptools import setup, find_packages
from os import listdir

with open('requirements.txt', 'r') as f:
    dependencies = f.read().splitlines()

with open("README.rst") as f:
    long_desc = f.read()

descr = "Chainladder Package - P&C Loss Reserving package "
name = 'chainladder'
url = 'https://github.com/casact/chainladder-python'
version='0.8.13' # Put this in __init__.py

data_path = ''
setup(
    name=name,
    version=version,
    maintainer='John Bogaardt',
    maintainer_email='jbogaardt@gmail.com',
    packages=find_packages(include=["chainladder", "chainladder.*"]),
    scripts=[],
    url=url,
    download_url='{}/archive/v{}.tar.gz'.format(url, version),
    license='MPL-2.0',
    include_package_data=True,
    package_data={
        'data': [data_path + item
                 for item in listdir('chainladder{}'.format(data_path))]},
    description=descr,
    long_description=long_desc,
    install_requires=dependencies,
)
